/***************************************************************************
                          designer.cpp  -  description
                             -------------------
    begin                : Wed May 24 16:16:04 EDT 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>
// include files for QT
#include <qdir.h>
#include <qstrlist.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qclipboard.h>

// include files for KDE
#include <kiconloader.h>
#include <kmsgbox.h>
#include <kfiledialog.h>

// application specific includes
#include "designer.h"
#include "designerview.h"
#include "designerdoc.h"
#include "resource.h"
#include "varwnd.h"
#include "outwnd.h"
#include "watchwnd.h"
#include "phpdbgr.h"
#include "./kwrite/kwview.h"
#include "cmapdlg.h"

static const char*close_xpm[]={
"5 5 2 1",
"# c black",
". c None",
"#...#",
".#.#.",
"..#..",
".#.#.",
"#...#"};


DesignerApp::DesignerApp() : KDockMainWindow( "MainFrame" )
{
	gpMainFrame = this;

  config=kapp->getConfig();

  ///////////////////////////////////////////////////////////////////
  // call inits to invoke all other construction parts
  initMenuBar();
  initToolBar();
  initStatusBar();
  initKeyAccel();
  initDocument();
	
  m_mainDock = createDockWidget( "MainView", QPixmap( close_xpm) );
  setView( m_mainDock );
  setMainDockWidget( m_mainDock );
  view = new DesignerView( m_mainDock );
  doc->addView(view);
	m_mainDock->setWidget( view );
	m_mainDock->setEnableDocking( KDockWidget::DockNone );

	m_dockVar = createDockWidget( "Variables", QPixmap( close_xpm ) );
  m_wndVar = new CVariablesWnd( m_dockVar, "Variables Window" );
  m_dockVar->setWidget( m_wndVar );

	m_dockOut = createDockWidget( "Output", QPixmap( close_xpm ) );
	m_wndOut = new COutputWnd( m_dockOut, "Output Window" );
	m_dockOut->setWidget( m_wndOut );

	m_dockWatch = createDockWidget( "Watch", QPixmap( close_xpm ) );
  m_wndWatch = new CWatchWnd( m_dockWatch, "Watch Window" );
  m_dockWatch->setWidget( m_wndWatch );

	config->setGroup( "WS - Editor" );
	if ( config->hasKey( "Version" ) ){
		readDockConfig( config, "WS - Editor" );
	}
	else{
		m_dockWatch->changeHideShowState();
		m_dockVar->changeHideShowState();
		m_dockOut->changeHideShowState();
	}

	m_phpdbgr = new CPHPDebugger( this );
	connect( doc, SIGNAL( notifyBPChange( const QString&, int ) ), m_phpdbgr, SLOT( slotOnBPChange( const QString& ) ) );
	connect( m_wndWatch, SIGNAL( notifyWatchListChanged() ), m_phpdbgr, SLOT( slotOnWatchListChange() ) );

  readOptions();

	//	Connect to the 'UpdateCommands' signal fron the app - our cue to update toolbar command status
	connect( kapp, SIGNAL( UpdateCommands() ), this, SLOT( slotUpdateToolbar() ) );
	connect( kapp, SIGNAL( UpdateCommands() ), this, SLOT( slotUpdateStatusBar() ) );

  connect( view,  SIGNAL(writeSelected(int)), doc,  SLOT(selectWrite(int)));

}

DesignerApp::~DesignerApp()
{
	doc = NULL;
}

void DesignerApp::initKeyAccel()
{
  keyAccel = new KAccel(this);
	
  // fileMenu accelerators
  keyAccel->connectItem(KAccel::New, this, SLOT(slotFileNew()));
  keyAccel->connectItem(KAccel::Open, this, SLOT(slotFileOpen()));
  keyAccel->connectItem(KAccel::Save, this, SLOT(slotFileSave()));
  keyAccel->connectItem(KAccel::Close, this, SLOT(slotFileClose()));
  keyAccel->connectItem(KAccel::Print, this, SLOT(slotFilePrint()));
  keyAccel->connectItem(KAccel::Quit, this, SLOT(slotFileQuit()));
  // editMenu accelerators
	keyAccel->connectItem(KAccel::Undo, this, SLOT(slotEditUndo()));
  keyAccel->connectItem(KAccel::Cut, this, SLOT(slotEditCut()));
  keyAccel->connectItem(KAccel::Copy, this, SLOT(slotEditCopy()));
  keyAccel->connectItem(KAccel::Paste, this, SLOT(slotEditPaste()));
	keyAccel->connectItem(KAccel::Find, this, SLOT(slotEditFind()));
	keyAccel->connectItem(KAccel::Replace, this, SLOT(slotEditReplace()));

  keyAccel->connectItem(KAccel::Help, kapp, SLOT(appHelpActivated()));

	keyAccel->insertItem( i18n("Debug Go"), "Debug Go", "F5" );
	keyAccel->connectItem( "Debug Go", this, SLOT( slotDebugGo() ) );
	keyAccel->insertItem( i18n("Debug Step Over"), "Debug StepOver", "F10" );
	keyAccel->connectItem( "Debug StepOver", this, SLOT( slotDebugStepOver() ) );
	keyAccel->insertItem( i18n("Debug Step Into"), "Debug StepInto", "F11" );
	keyAccel->connectItem( "Debug StepInto", this, SLOT( slotDebugStepInto() ) );
	keyAccel->insertItem( i18n("Debug Run To Cursor"), "Debug RunToCursor", "F7" );
	keyAccel->connectItem( "Debug RunToCursor", this, SLOT( slotDebugRunToCursor() ) );
	keyAccel->insertItem( i18n("Find Next"), "FindNext", "F3" );
	keyAccel->connectItem( "FindNext", this, SLOT( slotEditFindNext() ) );
	keyAccel->insertItem( i18n("Redo"), "Redo", "CTRL+Y" );
	keyAccel->connectItem("Redo", this, SLOT(slotEditRedo()));
	keyAccel->insertItem( i18n("Indent"), "Indent", "CTRL+I" );
	keyAccel->connectItem("Indent", this, SLOT(slotEditIndent()));
	keyAccel->insertItem( i18n("Unindent"), "Unindent", "CTRL+U" );
	keyAccel->connectItem("Unindent", this, SLOT( slotEditUnindent()));

  keyAccel->changeMenuAccel(fileMenu, ID_FILE_NEW, KAccel::New);
  keyAccel->changeMenuAccel(fileMenu, ID_FILE_OPEN, KAccel::Open);
  keyAccel->changeMenuAccel(fileMenu, ID_FILE_SAVE, KAccel::Save);
  keyAccel->changeMenuAccel(fileMenu, ID_FILE_CLOSE, KAccel::Close);
  keyAccel->changeMenuAccel(fileMenu, ID_FILE_PRINT, KAccel::Print);
  keyAccel->changeMenuAccel(fileMenu, ID_FILE_QUIT, KAccel::Quit);

	keyAccel->changeMenuAccel(editMenu, ID_EDIT_UNDO, KAccel::Undo);
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_REDO, "Redo");
  keyAccel->changeMenuAccel(editMenu, ID_EDIT_CUT, KAccel::Cut);
  keyAccel->changeMenuAccel(editMenu, ID_EDIT_COPY, KAccel::Copy);
  keyAccel->changeMenuAccel(editMenu, ID_EDIT_PASTE, KAccel::Paste);
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_FIND, KAccel::Find );
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_FINDNEXT, "FindNext" );
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_REPLACE, KAccel::Replace );
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_INDENT, "Indent" );
	keyAccel->changeMenuAccel(editMenu, ID_EDIT_UNINDENT, "Unindent" );

	keyAccel->changeMenuAccel(debugMenu, ID_DEBUG_GO, "Debug Go" );
	keyAccel->changeMenuAccel(debugMenu, ID_DEBUG_STEPOVER, "Debug StepOver" );
	keyAccel->changeMenuAccel(debugMenu, ID_DEBUG_STEPINTO, "Debug StepInto" );
	keyAccel->changeMenuAccel(debugMenu, ID_DEBUG_RUNTOCURSOR, "Debug RunToCursor" );

  keyAccel->readSettings();	
}

void DesignerApp::initMenuBar()
{
  ///////////////////////////////////////////////////////////////////
  // MENUBAR
  recentFilesMenu = new QPopupMenu();
  connect(recentFilesMenu, SIGNAL(activated(int)), SLOT(slotFileOpenRecent(int)));
	connect(recentFilesMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

  ///////////////////////////////////////////////////////////////////
  // menuBar entry fileMenu
  fileMenu = new QPopupMenu();
  fileMenu->insertItem(Icon("filenew.xpm"), i18n("&New"), ID_FILE_NEW);
  fileMenu->insertItem(Icon("fileopen.xpm"), i18n("&Open..."), ID_FILE_OPEN);
  fileMenu->insertItem(i18n("Open &recent"), recentFilesMenu, ID_FILE_OPEN_RECENT);
  fileMenu->insertItem(i18n("&Close"), ID_FILE_CLOSE);
  fileMenu->insertSeparator();
  fileMenu->insertItem(Icon("filefloppy.xpm") ,i18n("&Save"), ID_FILE_SAVE);
  fileMenu->insertItem(i18n("Save &As..."), ID_FILE_SAVE_AS);
	fileMenu->insertItem(i18n("Save A&ll"), ID_FILE_SAVE_ALL);
  fileMenu->insertSeparator();
  fileMenu->insertItem(Icon("fileprint.xpm"), i18n("&Print..."), ID_FILE_PRINT);
  fileMenu->insertSeparator();
  fileMenu->insertItem(i18n("E&xit"), ID_FILE_QUIT);
	
  ///////////////////////////////////////////////////////////////////
  // menuBar entry editMenu
  editMenu = new QPopupMenu();
	editMenu->insertItem(Icon("undo.xpm"), i18n("&Undo"), ID_EDIT_UNDO );
	editMenu->insertItem(Icon("redo.xpm"), i18n("&Redo"), ID_EDIT_REDO );
	editMenu->insertSeparator();
  editMenu->insertItem(Icon("editcut.xpm"), i18n("Cu&t"), ID_EDIT_CUT);
  editMenu->insertItem(Icon("editcopy.xpm"), i18n("&Copy"), ID_EDIT_COPY);
  editMenu->insertItem(Icon("editpaste.xpm"), i18n("&Paste"), ID_EDIT_PASTE);
	editMenu->insertSeparator();
	editMenu->insertItem(Icon("find.xpm"), i18n( "Find..." ), ID_EDIT_FIND );
	editMenu->insertItem( i18n( "Find Next" ), ID_EDIT_FINDNEXT );
	editMenu->insertItem( i18n( "Replace..." ), ID_EDIT_REPLACE );
	editMenu->insertSeparator();
	editMenu->insertItem( i18n( "&Indent" ), ID_EDIT_INDENT );
	editMenu->insertItem( i18n( "&Unindent" ), ID_EDIT_UNINDENT );

  ///////////////////////////////////////////////////////////////////
  // menuBar entry viewMenu
  viewMenu = new QPopupMenu();
  viewMenu->setCheckable(true);
  viewMenu->insertItem(i18n("&Toolbar"), ID_VIEW_TOOLBAR);
  viewMenu->insertItem(i18n("&Statusbar"), ID_VIEW_STATUSBAR);
	viewMenu->insertSeparator();

	QPopupMenu* optionsMenu = new QPopupMenu();
	optionsMenu->insertItem( i18n( "&Servers..." ), ID_OPTIONS_SERVERS );
	optionsMenu->insertSeparator();
	optionsMenu->insertItem( i18n( "&Editor..." ), ID_OPTIONS_EDITOR );
	optionsMenu->insertItem( i18n( "Editor Colors..." ), ID_OPTIONS_EDITOR_COLORS );
	optionsMenu->insertItem( i18n( "Editor Defaults..." ), ID_OPTIONS_EDITOR_DEFAULTS );
	optionsMenu->insertItem( i18n( "Syntax Highlighting..." ), ID_OPTIONS_SYNTAX_HIGHLIGHTING );

	///////////////////////////////////////////////////////////////////
	debugMenu = new QPopupMenu();
	debugMenu->insertItem( i18n( "&Options..." ), ID_DEBUG_OPTIONS );
	debugMenu->insertItem( i18n( "&Go" ), ID_DEBUG_GO );
	debugMenu->insertItem( i18n( "&Step Over" ), ID_DEBUG_STEPOVER );
	debugMenu->insertItem( i18n( "Step &Into" ), ID_DEBUG_STEPINTO );
	debugMenu->insertItem( i18n( "Run to &Cursor" ), ID_DEBUG_RUNTOCURSOR );
	debugMenu->insertItem( i18n( "Sto&p" ), ID_DEBUG_STOP );
	debugMenu->insertItem( i18n( "Disassembly" ), ID_DEBUG_DISASM );

  ///////////////////////////////////////////////////////////////////
  // menuBar entry helpMenu
  QString aboutstring=kapp->appName()+" "+VERSION+"\n\n";

  helpMenu = new QPopupMenu();
  helpMenu = kapp->getHelpMenu(true, aboutstring);

  ///////////////////////////////////////////////////////////////////
  // MENUBAR CONFIGURATION
  // insert your popup menus with the according menubar entries in the order
  // they will appear later from left to right
  menuBar()->insertItem(i18n("&File"), fileMenu);
  menuBar()->insertItem(i18n("&Edit"), editMenu);
	menuBar()->insertItem(i18n("&Options"), optionsMenu);
	menuBar()->insertItem(i18n("&Debug"), debugMenu );
  menuBar()->insertItem(i18n("&View"), viewMenu);

  menuBar()->insertSeparator();
  menuBar()->insertItem(i18n("&Help"), helpMenu);

  ///////////////////////////////////////////////////////////////////
  // CONNECT THE MENU SLOTS WITH SIGNALS
  // for execution slots and statusbar messages

  connect(fileMenu, SIGNAL(activated(int)), SLOT(commandCallback(int)));
  connect(fileMenu, SIGNAL(highlighted(int)), SLOT(statusCallback(int)));
	connect(fileMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

  connect(editMenu, SIGNAL(activated(int)), SLOT(commandCallback(int)));
  connect(editMenu, SIGNAL(highlighted(int)), SLOT(statusCallback(int)));
	connect(editMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

  connect(debugMenu, SIGNAL(activated(int)), SLOT(commandCallback(int)));
  connect(debugMenu, SIGNAL(highlighted(int)), SLOT(statusCallback(int)));
	connect(debugMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

  connect(optionsMenu, SIGNAL(activated(int)), SLOT(commandCallback(int)));
  connect(optionsMenu, SIGNAL(highlighted(int)), SLOT(statusCallback(int)));
	connect(optionsMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

  connect(viewMenu, SIGNAL(activated(int)), SLOT(commandCallback(int)));
  connect(viewMenu, SIGNAL(highlighted(int)), SLOT(statusCallback(int)));
	connect(viewMenu, SIGNAL(aboutToShow()), SLOT(slotUpdateMenu()));

}

void DesignerApp::initToolBar()
{

  ///////////////////////////////////////////////////////////////////
  // TOOLBAR
  toolBar()->insertButton(Icon("filenew.xpm"), ID_FILE_NEW, true, i18n("New File"));
  toolBar()->insertButton(Icon("fileopen.xpm"), ID_FILE_OPEN, true, i18n("Open File"));
  toolBar()->insertButton(Icon("filefloppy.xpm"), ID_FILE_SAVE, true, i18n("Save File"));
  toolBar()->insertButton(Icon("fileprint.xpm"), ID_FILE_PRINT, true, i18n("Print"));
  toolBar()->insertSeparator();
  toolBar()->insertButton(Icon("editcut.xpm"), ID_EDIT_CUT, true, i18n("Cut"));
  toolBar()->insertButton(Icon("editcopy.xpm"), ID_EDIT_COPY, true, i18n("Copy"));
  toolBar()->insertButton(Icon("editpaste.xpm"), ID_EDIT_PASTE, true, i18n("Paste"));
  toolBar()->insertSeparator();
  toolBar()->insertButton(Icon("help.xpm"), ID_HELP_CONTENTS, SIGNAL(clicked()),
  				kapp, SLOT(appHelpActivated()), true,i18n("Help"));

  ///////////////////////////////////////////////////////////////////
  // INSERT YOUR APPLICATION SPECIFIC TOOLBARS HERE WITH toolBar(n)


  ///////////////////////////////////////////////////////////////////
  // CONNECT THE TOOLBAR SLOTS WITH SIGNALS - add new created toolbars by their according number
  // connect for invoking the slot actions
  connect(toolBar(), SIGNAL(clicked(int)), SLOT(commandCallback(int)));
  // connect for the status help on holing icons pressed with the mouse button
  connect(toolBar(), SIGNAL(pressed(int)), SLOT(statusCallback(int)));

}

void DesignerApp::initStatusBar()
{
  ///////////////////////////////////////////////////////////////////
  // STATUSBAR
  statusBar()->insertItem(i18n("Line: 00000 Col: 000"), ID_STATUS_CLM);
  statusBar()->changeItem("", ID_STATUS_CLM);

  statusBar()->insertItem(i18n(IDS_STATUS_DEFAULT), ID_STATUS_MSG);

  statusBar()->setInsertOrder(KStatusBar::RightToLeft);
}

void DesignerApp::initDocument()
{
  doc = new DesignerDoc(this);
}

void DesignerApp::initView()
{ 
  ////////////////////////////////////////////////////////////////////
  // create the main widget here that is managed by KTMainWindow's view-region and
  // connect the widget to your document to display document contents.

  view = new DesignerView(this);
  doc->addView(view);
  setView(view);
  QString caption=kapp->getCaption();	
  setCaption(caption+": "+doc->getTitle());

}

void DesignerApp::addRecentFile(const QString &file)
{
  if(recentFiles.find(file) == -1)
  {
    if( recentFiles.count() < 5)
    {
      recentFiles.insert(0, file);
    }
    else
    {
      recentFiles.remove(4);
      recentFiles.insert(0, file);
    }
    recentFilesMenu->clear();
    for ( int i=0 ; i < (int) recentFiles.count(); i++)
    {
      recentFilesMenu->insertItem(recentFiles.at(i));
    }
  }
}

void DesignerApp::openDocumentFile(const char* _cmdl)
{
  slotStatusMsg(i18n("Opening file..."));

  doc->openDocument(_cmdl);
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}


DesignerDoc *DesignerApp::getDocument() const
{
  return doc;
}

void DesignerApp::saveOptions()
{	
  config->setGroup("General Options");
  config->writeEntry("Geometry", size());
  config->writeEntry("Show Toolbar", toolBar()->isVisible());
  config->writeEntry("Show Statusbar",statusBar()->isVisible());
  config->writeEntry("MenuBarPos", (int) menuBar()->menuBarPos());
  config->writeEntry("ToolBarPos", (int) toolBar()->barPos());
  config->writeEntry("Recent Files", recentFiles);
}


void DesignerApp::readOptions()
{
	
  config->setGroup("General Options");

  // bar status settings
  bool bViewToolbar = config->readBoolEntry("Show Toolbar", true);
  viewMenu->setItemChecked(ID_VIEW_TOOLBAR, bViewToolbar);
  if(!bViewToolbar)
  {
     enableToolBar(KToolBar::Hide);
  }
	
  bool bViewStatusbar = config->readBoolEntry("Show Statusbar", true);
  viewMenu->setItemChecked(ID_VIEW_STATUSBAR, bViewStatusbar);
  if(!bViewStatusbar)
  {
    enableStatusBar(KStatusBar::Hide);
  }

  // bar position settings
  KMenuBar::menuPosition menuBarPos;
  menuBarPos=(KMenuBar::menuPosition) config->readNumEntry("MenuBarPos", KMenuBar::Top);

  KToolBar::BarPosition toolBarPos;
  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("ToolBarPos", KToolBar::Top);

  menuBar()->setMenuBarPos(menuBarPos);
  toolBar()->setBarPos(toolBarPos);
	
  // initialize the recent file list
  recentFiles.setAutoDelete(TRUE);
  config->readListEntry("Recent Files", recentFiles);
	
  for (int i=0; i < (int) recentFiles.count(); i++)
  {
    recentFilesMenu->insertItem(recentFiles.at(i));
  }

  QSize size=config->readSizeEntry("Geometry");
  if(!size.isEmpty())
  {
    resize(size);
  }
}

void DesignerApp::saveProperties(KConfig *_cfg)
{
  if(doc->getTitle()!=i18n("Untitled") && !doc->isModified())
  {
    // saving to tempfile not necessary

  }
  else
  {
    QString filename=doc->getAbsFilePath();	
    _cfg->writeEntry("filename", filename);
    _cfg->writeEntry("modified", doc->isModified());
		
    QString tempname = kapp->tempSaveName(filename);
    doc->saveDocument(tempname);
  }
}


void DesignerApp::readProperties(KConfig* _cfg)
{
  QString filename = _cfg->readEntry("filename", "");
  bool modified = _cfg->readBoolEntry("modified", false);
  if(modified)
  {
    bool canRecover;
    QString tempname = kapp->checkRecoverFile(filename, canRecover);
  	
    if(canRecover)
    {
      doc->openDocument(tempname);
      doc->setModified();
      QFileInfo info(filename);
      doc->setAbsFilePath(info.absFilePath());
      doc->setTitle(info.fileName());
      QFile::remove(tempname);
    }
  }
  else
  {
    if(!filename.isEmpty())
    {
      doc->openDocument(filename);
    }
  }

  QString caption=kapp->getCaption();	
  setCaption(caption+": "+doc->getTitle());
}		

bool DesignerApp::queryClose()
{
  return doc->saveModified();
}

bool DesignerApp::queryExit()
{
  saveOptions();
  return true;
}

/////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATION
/////////////////////////////////////////////////////////////////////
void DesignerApp::slotFileNew()
{
  slotStatusMsg(i18n("Creating new document..."));

  doc->newDocument();		

  QString caption=kapp->getCaption();	
  setCaption(caption+": "+doc->getTitle());

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFileOpen()
{
  slotStatusMsg(i18n("Opening file..."));

  QString fileToOpen = KFileDialog::getOpenFileName( QDir::homeDirPath(),
    i18n("*.*htm* *.php* *.asp *.cfm *.css *.jss *.js *.inc* *.*HTM* *.PHP* *.ASP *.CFM *.CSS *.JSS *.JS *.INC*|HTML files\n*|All files" ),
    this, i18n("Open File..."));

	if( !fileToOpen.isEmpty() ){
		doc->openDocument(fileToOpen);
		QString caption = kapp->getCaption();
		setCaption( caption + ": " + doc->getTitle() );
		addRecentFile( fileToOpen );
	}

  slotStatusMsg( i18n(IDS_STATUS_DEFAULT) );
}

void DesignerApp::slotFileOpenRecent(int id_)
{
  slotStatusMsg(i18n("Opening file..."));
	
  doc->openDocument(recentFiles.at(id_));
  QString caption=kapp->getCaption();	
  setCaption(caption+": "+doc->getTitle());

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFileSave()
{
  slotStatusMsg(i18n("Saving file..."));

  if ( !doc->write->hasFileName() ){
  	slotFileSaveAs();
	}
  else{
  	doc->saveDocument(doc->getAbsFilePath());
	}

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFileSaveAs()
{
  slotStatusMsg(i18n("Saving file with a new filename..."));

  QString newName=KFileDialog::getSaveFileName(QDir::currentDirPath(),
                                               i18n("*|All files"), this, i18n("Save as..."));
  if(!newName.isEmpty())
  {
    doc->write->setModified( true);


    QFileInfo saveAsInfo(newName);
    doc->setTitle(saveAsInfo.fileName());
    doc->setAbsFilePath(saveAsInfo.absFilePath());
    doc->write->setModified( true);
    doc->saveDocument(newName);
    addRecentFile(newName);

    QString caption=kapp->getCaption();	
    setCaption(caption+": "+doc->getTitle());
  }

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFileClose()
{
  slotStatusMsg(i18n("Closing file..."));
	
  if( doc->write->canDiscard() )
  {
    doc->closeDocument();

    QString caption=kapp->getCaption();	
    setCaption(caption+": "+doc->getTitle());
  }

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFilePrint()
{
  slotStatusMsg(i18n("Printing..."));

  QPrinter printer;
  if (printer.setup(this))
  {
    view->print(&printer);
  }

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotFileQuit()
{
  slotStatusMsg(i18n("Exiting..."));
  saveOptions();
  // close the first window, the list makes the next one the first again.
  // This ensures that queryClose() is called on each window to ask for closing
  KTMainWindow* w;
  if(memberList)
  {
    for(w=memberList->first(); w!=0; w=memberList->first())
    {
      // only close the window if the closeEvent is accepted. If the user presses Cancel on the saveModified() dialog,
      // the window and the application stay open.
      if(!w->close())
	break;
    }
  }	
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditCut()
{
  slotStatusMsg(i18n("Cutting selection..."));
	doc->write->cut();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditCopy()
{
  slotStatusMsg(i18n("Copying selection to clipboard..."));
	doc->write->copy();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditPaste()
{
  slotStatusMsg(i18n("Inserting clipboard contents..."));
	doc->write->paste();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotViewToolBar()
{
  slotStatusMsg(i18n("Toggle the toolbar..."));
  ///////////////////////////////////////////////////////////////////
  // turn Toolbar on or off
  if( viewMenu->isItemChecked(ID_VIEW_TOOLBAR))
  {
    viewMenu->setItemChecked(ID_VIEW_TOOLBAR, false);
    enableToolBar(KToolBar::Hide);
  }
  else
  {
    viewMenu->setItemChecked(ID_VIEW_TOOLBAR, true);
    enableToolBar(KToolBar::Show);
  }		

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotViewStatusBar()
{
  slotStatusMsg(i18n("Toggle the statusbar..."));
  ///////////////////////////////////////////////////////////////////
  //turn Statusbar on or off
  if( viewMenu->isItemChecked(ID_VIEW_STATUSBAR))
  {
    viewMenu->setItemChecked(ID_VIEW_STATUSBAR, false);
    enableStatusBar(KStatusBar::Hide);
  }
  else
  {
    viewMenu->setItemChecked(ID_VIEW_STATUSBAR, true);
    enableStatusBar(KStatusBar::Show);
  }

  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}


void DesignerApp::slotStatusMsg(const QString &text)
{
  ///////////////////////////////////////////////////////////////////
  // change status message permanently
  statusBar()->clear();
  statusBar()->changeItem(text, ID_STATUS_MSG);
}


void DesignerApp::slotStatusHelpMsg(const QString &text)
{
  ///////////////////////////////////////////////////////////////////
  // change status message of whole statusbar temporary (text, msec)
  statusBar()->message(text, 2000);
}


void DesignerApp::commandCallback(int id_)
{
	if ( !( commandEnableCallback( id_ ) & CMDUI_ENABLE ) ){
		return;
	}

  switch (id_)
  {
    case ID_FILE_NEW:
    	 slotFileNew();
         break;

    case ID_FILE_OPEN:
         slotFileOpen();
         break;

    case ID_FILE_SAVE:
         slotFileSave();
         break;

    case ID_FILE_SAVE_AS:
         slotFileSaveAs();
         break;

    case ID_FILE_CLOSE:
         slotFileClose();
         break;

    case ID_FILE_PRINT:
         slotFilePrint();
         break;

    case ID_FILE_QUIT:
         slotFileQuit();
         break;

		case ID_EDIT_UNDO:
				slotEditUndo();
				break;

		case ID_EDIT_REDO:
				slotEditRedo();
				break;

    case ID_EDIT_CUT:
         slotEditCut();
         break;

    case ID_EDIT_COPY:
         slotEditCopy();
         break;

    case ID_EDIT_PASTE:
         slotEditPaste();
         break;
 
		case ID_EDIT_FIND:
				slotEditFind();
				break;

		case ID_EDIT_FINDNEXT:
				slotEditFindNext();
				break;

		case ID_EDIT_REPLACE:
				slotEditReplace();
				break;

    case ID_VIEW_TOOLBAR:
         slotViewToolBar();
         break;

    case ID_VIEW_STATUSBAR:
         slotViewStatusBar();
         break;

		case ID_DEBUG_GO:
  			 slotDebugGo();
				 break;

		case ID_DEBUG_STEPOVER:
  			 slotDebugStepOver();
				 break;

		case ID_DEBUG_STEPINTO:
  			 slotDebugStepInto();
				 break;

		case ID_DEBUG_RUNTOCURSOR:
  			 slotDebugRunToCursor();
				 break;

		case ID_DEBUG_STOP:
  			 slotDebugStop();
				 break;

		case ID_DEBUG_DISASM:
  			 slotDebugDisasm();
				 break;

		case ID_DEBUG_OPTIONS:
//				 slotDebugOptions();
				 break;

		case ID_OPTIONS_SERVERS:
			slotOptionsServers();
			break;

	  case ID_OPTIONS_EDITOR:
			slotOptionsEditor();
      break;

		case ID_OPTIONS_EDITOR_DEFAULTS:
			slotOptionsEditorDefaults();
			break;

		case ID_OPTIONS_EDITOR_COLORS:
			slotOptionsEditorColors();
			break;

		case ID_OPTIONS_SYNTAX_HIGHLIGHTING:
			slotOptionsSyntaxHighlighting();
			break;

    default:
         break;
  }
	if ( ID_VIEW_HIGHLIGHT <= id_ && id_ <= ID_VIEW_HIGHLIGHT_LAST ){
		slotViewHighlight( id_ - ID_VIEW_HIGHLIGHT );
	}
}

void DesignerApp::statusCallback(int id_)
{
  switch (id_)
  {
    case ID_FILE_NEW_WINDOW:
         slotStatusHelpMsg(i18n("Opens a new application window"));
         break;

    case ID_FILE_NEW:
         slotStatusHelpMsg(i18n("Creates a new document"));
         break;

    case ID_FILE_OPEN:
         slotStatusHelpMsg(i18n("Opens an existing document"));
         break;

    case ID_FILE_OPEN_RECENT:
         slotStatusHelpMsg(i18n("Opens a recently used file"));
         break;

    case ID_FILE_SAVE:
         slotStatusHelpMsg(i18n("Saves the actual document"));
         break;

    case ID_FILE_SAVE_AS:
         slotStatusHelpMsg(i18n("Saves the actual document as..."));
         break;

    case ID_FILE_CLOSE:
         slotStatusHelpMsg(i18n("Closes the actual document"));
         break;

    case ID_FILE_PRINT:
         slotStatusHelpMsg(i18n("Prints out the actual document"));
         break;

    case ID_FILE_QUIT:
         slotStatusHelpMsg(i18n("Quits the application"));
         break;

    case ID_EDIT_CUT:
         slotStatusHelpMsg(i18n("Cuts the selected section and puts it to the clipboard"));
         break;

    case ID_EDIT_COPY:
         slotStatusHelpMsg(i18n("Copies the selected section to the clipboard"));
         break;

    case ID_EDIT_PASTE:
         slotStatusHelpMsg(i18n("Pastes the clipboard contents to actual position"));
         break;

    case ID_VIEW_TOOLBAR:
         slotStatusHelpMsg(i18n("Enables/disables the toolbar"));
         break;

    case ID_VIEW_STATUSBAR:
         slotStatusHelpMsg(i18n("Enables/disables the statusbar"));
         break;

    default:
         break;
  }
}

int DesignerApp::commandEnableCallback( int id_ )
{
	int rc = CMDUI_ENABLE;
	switch ( id_ ){
	case ID_FILE_SAVE:
	case ID_FILE_SAVE_AS:
		if ( !doc->isModified() ){
			rc = 0;
		}
		break;
	case ID_EDIT_UNDO:
		if ( 0 != doc->write && !( doc->write->undoState() & 1 ) ){
			rc = 0;
		}
		break;
	case ID_EDIT_REDO:
		if ( 0 != doc->write && !(doc->write->undoState() & 2 ) ){
			rc = 0;
		}
		break;			
	case ID_EDIT_CUT:
  case ID_EDIT_COPY:
		if ( 0 != doc->write && doc->write->markedText().isEmpty() ){
			rc = 0;
		}
		break;
  case ID_EDIT_PASTE:
    if ( 0 == QApplication::clipboard()->text() ){
			rc = 0;
		}
		break;
	case ID_DEBUG_STEPOVER:
	case ID_DEBUG_STEPINTO:
	case ID_DEBUG_RUNTOCURSOR:
	case ID_DEBUG_STOP:
	case ID_DEBUG_DISASM:
		if ( 0 == pdbgr() || !pdbgr()->fIsActive() ){
			rc = 0;
		}
	}
	return rc;		
}

CDebugger* DesignerApp::pdbgr()
{
	return m_phpdbgr;
}

QString		DesignerApp::PrepareWatchList()
{
	return m_wndWatch->PrepareWatchList();
}

void			DesignerApp::PrepareVariableList()
{
	m_wndVar->Prepare();
}

void			DesignerApp::AddVariable( int i, const char* pcszParent, const char* pcszName, const char* pcszValue )
{
	m_wndVar->AddVar( i, pcszParent, pcszName, pcszValue );
}

void			DesignerApp::AddWatch( const char* pcszParent, const char* pcszName, const char* pcszValue )
{
	m_wndWatch->AddVar( pcszParent, pcszName, pcszValue );
}

void			DesignerApp::PurgeVariableList()
{
	m_wndVar->Purge();
}

void			DesignerApp::PurgeWatchList()
{
	m_wndWatch->Purge();
}

void			DesignerApp::AppendStdOut( const char* pcsz )
{
	m_wndOut->Append( COutputWnd::OW_STDOUT, pcsz );
}

void			DesignerApp::AppendStdErr( const char* pcsz )
{
	m_wndOut->Append( COutputWnd::OW_STDERR, pcsz );
}

void			DesignerApp::AppendTranscript( const char* pcsz )
{
	m_wndOut->Append( COutputWnd::OW_TRANSCRIPT, pcsz );
}

void			DesignerApp::ClearDbgWnds()
{
	m_wndOut->Clear();
	m_wndVar->Prepare();
	m_wndVar->Purge();
	m_wndWatch->PrepareWatchList();
	m_wndWatch->Purge();
}

void			DesignerApp::setStepLine( const char* pcszServer, const char* pcszFileName, int iLine )
{
	QString strLocal = mapFileName( pcszServer, pcszFileName );
	if ( 0 == strcmp( doc->getAbsFilePath(), strLocal ) ){
		doc->setStepLine( iLine );
	}
	else{
		doc->setStepLine( -1 );
		if ( doc->openDocument( strLocal ) ){
			doc->setStepLine( iLine );
			QString caption = kapp->getCaption();
			setCaption( caption + ": " + doc->getTitle() );
			addRecentFile( strLocal );
		}	
	}
}

//	Convert '\' to '/' and add a trailing for '/' for directories
QString		DesignerApp::normalizeFileName( const char* pcszFileName, bool fIsDirectory )
{
	QString str( pcszFileName );
	int i = str.find( '\\' );
	while ( -1 != i ){
		str.setExpand( i, '/' );
		i = str.find( '\\', i );
	}
	if ( fIsDirectory && str.right( 1 ) != "/" ){
		str += "/";
	}
	return str;
}

//	Given a server-based file name, uses the server mappings to find the corresponding local file
QString		DesignerApp::mapFileName( const char* pcszServer, const char* pcszServerPath )
{
	KConfig* config = kapp->getConfig();
	config->setGroup( "Mapping" );

	QString strServerPath = normalizeFileName( pcszServerPath );
	QString strRet( strServerPath );

	int cBestSoFar = 0;
	int i = 0;
	char szName[ 15 ];
	sprintf( szName, "Map%i", i );
	QStrList lststr;
	while ( config->readListEntry( szName, lststr ) ){
		if ( 0 == strcmp( lststr.at( 0 ), "" ) ){
			break;
		}
		if ( 0 == stricmp( lststr.at( 0 ), pcszServer ) ){
			QString strServerTry = normalizeFileName( lststr.at( 2 ), true );
			int cTry = strServerTry.length();
			if ( cBestSoFar < cTry && 0 == strnicmp( strServerTry, strServerPath, cTry ) ){
				cBestSoFar = cTry;
				strRet = normalizeFileName( lststr.at( 1 ), true ) +
					normalizeFileName( strServerPath.mid( cTry, strServerPath.length() - cTry ) );
			}
		}
		++i;
		sprintf( szName, "Map%i", i );
	}

	return strRet;
}

QString	DesignerApp::getBreakpoints( const char* pcszServer, const char* pcszFileName )
{
	QString strLocal = mapFileName( pcszServer, pcszFileName );
	return doc->getBreakpoints( strLocal );
}

void 			DesignerApp::slotDebugGo()
{
	if ( !pdbgr()->fIsActive() ){
		KConfig* config = kapp->getConfig();
		writeDockConfig( config, "WS - Editor" );
		config->setGroup( "WS - Debug" );
		if ( config->hasKey( "Version" ) ){
			readDockConfig( config, "WS - Debug" );
		}
		else{
		  m_dockVar->manualDock( m_mainDock, KDockWidget::DockBottom, 40 );
			m_dockOut->manualDock( m_dockVar, KDockWidget::DockBottom, 50 );
			m_dockWatch->manualDock( m_dockVar, KDockWidget::DockRight, 50 );
		}
		pdbgr()->Activate();
	}
	pdbgr()->Go();
}

void 			DesignerApp::slotDebugStepOver()
{
	assert( pdbgr() && pdbgr()->fIsActive() );
	pdbgr()->StepOver();
}

void 			DesignerApp::slotDebugStepInto()
{
	assert( pdbgr() && pdbgr()->fIsActive() );
	pdbgr()->StepInto();
}

void 			DesignerApp::slotDebugRunToCursor()
{
}

void 			DesignerApp::slotDebugStop()
{
	assert( pdbgr() && pdbgr()->fIsActive() );
	pdbgr()->Stop();
	doc->setStepLine( -1 );
	writeDockConfig( kapp->getConfig(), "WS - Debug" );
	readDockConfig( kapp->getConfig(), "WS - Editor" );
}

void 			DesignerApp::slotDebugDisasm()
{
}
/*
void			DesignerApp::slotDebugOptions()
{
	CDbgFDlg	dlg( this );
	dlg.exec();
}
*/

//	Called in idle time to update toolbar button status
void 			DesignerApp::slotUpdateToolbar()
{
	if ( NULL == doc ){
		return;
	}
	int aid[] = { ID_FILE_SAVE, ID_FILE_SAVE_AS,
		ID_EDIT_CUT, ID_EDIT_COPY, ID_EDIT_PASTE };	//	Really ought to get this list by introspection, but KToolbars don't have it.
	for ( int i = 0 ; i < (int)(sizeof( aid ) / sizeof( aid[ 0 ] )) ; ++i ){
		int iCmdUI = commandEnableCallback( aid[ i ] );
		toolBar()->setItemEnabled( aid[ i ], iCmdUI & CMDUI_ENABLE );
	}
}

//	Also called in idle time to update panes in the status bar
void		DesignerApp::slotUpdateStatusBar()
{
  QString linenumber;
	if ( doc && doc->write ){
	  int y = view->write()->currentLine()+1;
	  int x = view->write()->currentColumn() +1;

	  linenumber.sprintf(i18n("Line: %d Col: %d"),y,x);
	}
	statusBar()->changeItem(linenumber.data(), ID_STATUS_CLM);
}

//	Called in response to QPopupMenu::aboutToShow to update menu status
void		DesignerApp::slotUpdateMenu()
{
	if ( sender() == viewMenu ){
		doc->updateViewMenu( viewMenu );
	}

	//	They don't give us the current popup in signal 'aboutToShow' so we refresh the whole menu bar.  Pah.
	int aid[] = { ID_FILE_SAVE, ID_FILE_SAVE_AS,
		ID_DEBUG_STEPOVER, ID_DEBUG_STEPINTO, ID_DEBUG_RUNTOCURSOR, ID_DEBUG_STOP, ID_DEBUG_DISASM,
		ID_EDIT_CUT, ID_EDIT_COPY, ID_EDIT_PASTE, ID_EDIT_UNDO, ID_EDIT_REDO
	};	//	Really ought to get this list by introspection...
	for ( int i = 0 ; i < (int)(sizeof( aid ) / sizeof( aid[ 0 ]) ) ; ++i ){
		int iCmdUI = commandEnableCallback( aid[ i ] );
		menuBar()->setItemEnabled( aid[ i ], iCmdUI & CMDUI_ENABLE );
	}
}

void	DesignerApp::slotOptionsServers()
{
	CMapDlg	dlg( this );
	dlg.exec();
}

void DesignerApp::slotOptionsEditor()
{
  doc->write->optDlg();
  doc->write->writeConfig( kapp->getConfig() );
}

void DesignerApp::slotOptionsEditorDefaults()
{
	doc->write->hlDef();
}

void DesignerApp::slotOptionsEditorColors()
{
	doc->write->colDlg();
	doc->write->writeConfig( kapp->getConfig() );
}

void DesignerApp::slotOptionsSyntaxHighlighting()
{
	doc->write->hlDlg();
}

void DesignerApp::slotEditUndo()
{
	doc->write->undo();
}

void DesignerApp::slotEditRedo()
{
	doc->write->redo();
}

void DesignerApp::slotEditFind()
{
  doc->write->search();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditFindNext()
{
  doc->write->searchAgain();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditReplace()
{
  doc->write->replace();
  slotStatusMsg(i18n(IDS_STATUS_DEFAULT));
}

void DesignerApp::slotEditIndent()
{
	doc->write->indent();
}

void DesignerApp::slotEditUnindent()
{
	doc->write->unIndent();
}

void DesignerApp::slotViewHighlight( int index )
{
	doc->setWriteMode( index );
}
