/***************************************************************************
                          designer.h  -  description
                             -------------------
    begin                : Wed May 24 16:16:04 EDT 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef DESIGNER_H
#define DESIGNER_H
 

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


// include files for KDE 
#include <kapp.h>
#include <ktmainwindow.h>
#include <kaccel.h>

#include "kdockwidget.h"

class CAppWithIdle : public KApplication
{
	Q_OBJECT

public:
	CAppWithIdle( int argc, char* argv[], const char* pszApp, const char* pszHelpDir );
	~CAppWithIdle();
	bool notify( QObject* pobj, QEvent* pev );

public slots:
	void appHelpActivated();

signals:
	void UpdateCommands();

protected:
	QString m_strHelpDir;
};

// forward declaration of the Designer classes
class DesignerDoc;
class DesignerView;
class CDebugger;
class CVariablesWnd;
class	CWatchWnd;
class COutputWnd;

/**
  * The base class for Designer application windows. It sets up the main
  * window and reads the config file as well as providing a menubar, toolbar
  * and statusbar. An instance of OpenFusionView creates your center view, which is connected
  * to the window's Doc object.
  * DesignerApp reimplements the methods that KTMainWindow provides for main window handling and supports
  * full session management as well as keyboard accelerator configuration by using KAccel.
  * @see KTMainWindow
  * @see KApplication
  * @see KConfig
  * @see KAccel
  *
  * @author Source Framework Automatically Generated by KDevelop, (c) The KDevelop Team.
  * @version KDevelop version 0.4 code generation
  */
class DesignerApp : public KDockMainWindow
{
  Q_OBJECT

  friend class DesignerView;

  public:
    /** construtor of DesignerApp, calls all init functions to create the application.
     * @see initMenuBar initToolBar
     */
    DesignerApp();
    ~DesignerApp();

    /** add a opened file to the recent file list and update recent_file_menu
     */
    void addRecentFile(const QString &file);
    /** opens a file specified by commandline option
     */
    void openDocumentFile(const char *_cmdl);
    /** returns a pointer to the current document connected to the KTMainWindow instance and is used by
     * the View class to access the document object's methods
     */	
    DesignerDoc *getDocument() const; 	

		enum { CMDUI_ENABLE = 1 };

  protected:
    /** save general Options like all bar positions and status as well as the geometry and the recent file list to the configuration
     * file
     */ 	
    void saveOptions();
    /** read general Options again and initialize all variables like the recent file list
     */
    void readOptions();
    /** initKeyAccel creates the keyboard accelerator items for the available slots and changes the menu accelerators.
     * @see KAccel
     */
    void initKeyAccel();
    /** initMenuBar creates the menubar and inserts the menupopups as well as creating the helpMenu.
     * @see KApplication#getHelpMenu
     */
    void initMenuBar();
    /** this creates the toolbars.
     */
    void initToolBar();
    /** sets up the statusbar for the main window by initialzing a statuslabel.
     */
    void initStatusBar();
    /** initializes the document object of the main window that is connected to the view in initView().
     * @see initView();
     */
    void initDocument();
    /** creates the centerwidget of the KTMainWindow instance and sets it as the view
     */
    void initView();
    /** queryClose is called by KTMainWindow on each closeEvent of a window. Against the
     * default implementation (only returns true), this calles saveModified() on the document object to ask if the document shall
     * be saved if Modified; on cancel the closeEvent is rejected.
     * @see KTMainWindow#queryClose
     * @see KTMainWindow#closeEvent
     */
    virtual bool queryClose();
    /** queryExit is called by KTMainWindow when the last window of the application is going to be closed during the closeEvent().
     * Against the default implementation that just returns true, this calls saveOptions() to save the settings of the last window's	
     * properties.
     * @see KTMainWindow#queryExit
     * @see KTMainWindow#closeEvent
     */
    virtual bool queryExit();
    /** saves the window properties for each open window during session end to the session config file, including saving the currently
     * opened file by a temporary filename provided by KApplication.
     * @see KTMainWindow#saveProperties
     */
    virtual void saveProperties(KConfig *_cfg);
    /** reads the session config file and restores the application's state including the last opened files and documents by reading the
     * temporary files saved by saveProperties()
     * @see KTMainWindow#readProperties
     */
    virtual void readProperties(KConfig *_cfg);

  public slots:
    /** switch argument for slot selection by menu or toolbar ID */
    void commandCallback(int id_);
    /** switch argument for Statusbar help entries on slot selection. Add your ID's help here for toolbars and menubar entries. */
    void statusCallback(int id_);
		int commandEnableCallback( int id_ );
    /** clears the document in the actual view to reuse it as the new document */
    void slotFileNew();
    /** open a file and load it into the document*/
    void slotFileOpen();
    /** opens a file from the recent files menu */
    void slotFileOpenRecent(int id_);
    /** save a document */
    void slotFileSave();
    /** save a document by a new filename*/
    void slotFileSaveAs();
    /** asks for saving if the file is modified, then closes the actual file and window*/
    void slotFileClose();
    /** print the actual file */
    void slotFilePrint();
    /** closes all open windows by calling close() on each memberList item until the list is empty, then quits the application.
     * If queryClose() returns false because the user canceled the saveModified() dialog, the closing breaks.
     */
    void slotFileQuit();
		void slotEditUndo();
		void slotEditRedo();
		void slotEditIndent();
		void slotEditUnindent();

    /** put the marked text/object into the clipboard and remove
     *	it from the document
     */
    void slotEditCut();
    /** put the marked text/object into the clipboard
     */
    void slotEditCopy();
    /** paste the clipboard into the document
     */
    void slotEditPaste();
    /** toggles the toolbar
     */
    void slotViewToolBar();
    /** toggles the statusbar
     */
    void slotViewStatusBar();

		void slotViewHighlight( int i );

    /** changes the statusbar contents for the standard label permanently, used to indicate current actions.
     * @param text the text that is displayed in the statusbar
     */
    void slotStatusMsg(const QString &text);
    /** changes the status message of the whole statusbar for two seconds, then restores the last status. This is used to display
     * statusbar messages that give information about actions for toolbar icons and menuentries.
     * @param text the text that is displayed in the statusbar
     */
    void slotStatusHelpMsg(const QString &text);

		void slotUpdateToolbar();
		void slotUpdateMenu();
		void slotUpdateStatusBar();

		void slotEditFind();
		void slotEditFindNext();
		void slotEditReplace();

		void slotOptionsServers();
		void slotOptionsEditor();
		void slotOptionsEditorColors();
		void slotOptionsEditorDefaults();
		void slotOptionsSyntaxHighlighting();

		void slotDebugGo();
		void slotDebugStepOver();
		void slotDebugStepInto();
		void slotDebugRunToCursor();
		void slotDebugStop();
		void slotDebugDisasm();

  private:
    /** contains the recently used filenames */
    QStrList recentFiles;

    /** the configuration object of the application */
    KConfig *config;
    /** the key accelerator container */
    KAccel *keyAccel;
    /** file_menu contains all items of the menubar entry "File" */
    QPopupMenu *fileMenu;
    /** the recent file menu containing the last five opened files */
    QPopupMenu *recentFilesMenu;
    /** edit_menu contains all items of the menubar entry "Edit" */
    QPopupMenu *editMenu;
    /** debug_menu contains all items of the menubar entry "Debug" */
		QPopupMenu* debugMenu;
    /** view_menu contains all items of the menubar entry "View" */
    QPopupMenu *viewMenu;
    /** help_menu contains all items of the menubar entry "Help" */
    QPopupMenu *helpMenu;
    /** view is the main widget which represents your working area. The View
     * class should handle all events of the view widget.  It is kept empty so
     * you can create your view according to your application's needs by
     * changing the view class.
     */
    DesignerView *view;
    /** doc represents your actual document and is created only once. It keeps
     * information such as filename and does the serialization of your files.
     */
    DesignerDoc *doc;

//	Debugging
public:
	CDebugger*	pdbgr();

//	UI services provided for the debugger to use
	QString			PrepareWatchList();
	QString			getBreakpoints( const char* pcszServer, const char* pcszFileName );
	QString			mapFileName( const char* pcszServer, const char* pcszServerPath );

	void				PrepareVariableList();
	void				AddVariable( int i, const char* pcszParent, const char* pcszName, const char* pcszValue );
	void				AddWatch( const char* pcszParent, const char* pcszName, const char* pcszValue );
	void				PurgeVariableList();
	void				PurgeWatchList();
	void				AppendStdOut( const char* pcsz );
	void				AppendStdErr( const char* pcsz );
	void				AppendTranscript( const char* pcsz );
	void				ClearDbgWnds();
	void				setStepLine( const char* pcszServer, const char* pcszFileName, int iLine );

protected:
	QString			normalizeFileName( const char* pcszFileName, bool fIsDirectory = false );

	CDebugger*	m_phpdbgr;
	CVariablesWnd*	m_wndVar;
	CWatchWnd*			m_wndWatch;
	COutputWnd*			m_wndOut;

	KDockWidget*	m_mainDock;
	KDockWidget*	m_dockVar;
	KDockWidget*	m_dockWatch;
	KDockWidget*	m_dockOut;
};

extern DesignerApp* gpMainFrame;

#endif // DESIGNER_H
