/***************************************************************************
                          designerdoc.cpp  -  description
                             -------------------
    begin                : Wed May 24 16:16:04 EDT 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qdir.h>
#include <qfileinfo.h>
#include <qwidget.h>

// include files for KDE
#include <kapp.h>
#include <kmsgbox.h>

// application specific includes
#include "designerdoc.h"
#include "designer.h"
#include "designerview.h"

#include "kwrite/kwdoc.h"
#include "resource.h"

QList<DesignerView> *DesignerDoc::pViewList = 0L;

class COpenFileInfo
{
	
};

DesignerDoc::DesignerDoc(QWidget *parent, const char *name) : QObject(parent, name)
{
  if(!pViewList)
  {
    pViewList = new QList<DesignerView>();
  }

  fileList = new QDict<int>();
  fileNameList = new QStrList();
}

DesignerDoc::~DesignerDoc()
{
}

void DesignerDoc::addView(DesignerView *view)
{
  pViewList->append(view);
	write = view->write();
	this->view = view;
}

void DesignerDoc::removeView(DesignerView *view)
{
  pViewList->remove(view);
}

void DesignerDoc::setAbsFilePath(const QString &filename)
{
  write->setFileName(filename);
}

QString DesignerDoc::getAbsFilePath() const
{
   QString fname = write->fileName();
   if ( fname.left(5) == "file:" ) fname.remove(0,5);
   return fname;
}

void DesignerDoc::setTitle(const QString &_t)
{
  title=_t;
}

const QString &DesignerDoc::getTitle()
{
  QString fname = write->fileName();
  QFileInfo fileInfo(fname);
  title=(fileInfo.fileName()).data();
  return title;
}

void DesignerDoc::slotUpdateAllViews(DesignerView *sender)
{
  DesignerView *w;
  if(pViewList)
  {
    for(w=pViewList->first(); w!=0; w=pViewList->next())
    {
      if(w!=sender)
        w->repaint();
    }
  }

}

bool DesignerDoc::saveModified()
{
  return saveAll( false );
}

void DesignerDoc::closeDocument()
{
  QString fname = write->fileName();
  if ( !fname.data() ) fname = "Untitled.php";
  if ( fname.left(5) == "file:" ) fname.remove(0,5);

  fileList->remove( fname );
  fileNameList->remove( fileNameList->find( fname) );

  if ( !( write = view->removeWrite() ))
    newDocument();

  fname = write->fileName();
  QFileInfo fileInfo(fname);
  title=fileInfo.fileName();
}

bool DesignerDoc::newDocument( const char *filename = 0 )
{
  QString fname;
  bool fnew = false;
  if ( !filename ) {
    fname = "Untitled.php";
    fnew = true;
  } else fname = filename;

 	if ( !fileList->find( fname.data()) || fnew ) {
 	
 	  if ( write )
 	    if ( !write->hasFileName() && !write->isModified() && !fnew ) {
 	      view->removeWrite();
 	  }
 	
 	  newWrite( view);

    int *id = new int;
    *id = view->addWrite( write,fname.data());

    fileList->insert( fname.data(), id);
    fileNameList->insert( 0,fname );


    setModified(false);
    title = fname.data();


  } else {
    int id = *fileList->find( fname.data());
    view->fileTabBar->setCurrentTab( id );
    view->repaint();
    title = fname.data();
  }

  return true;
}

bool DesignerDoc::openDocument(const QString &filename, const char *format /*=0*/)
{
  QString fname = filename;
  if ( fname.left(5) == "file:" ) fname.remove(0,5);
  QFileInfo fileInfo(fname);

  if ( fileInfo.exists() ) {
    newDocument( fname.data() );

    title=fileInfo.fileName();

    write->loadURL(fname);
    write->setHl(write->getHl());
  }
  else return false;

  return true;
}

bool DesignerDoc::saveDocument(const QString &filename, const char *format /*=0*/)
{
  QFileInfo fileInfo(filename);
  title=fileInfo.fileName();

  changeFileName( filename );

  write->save();
  setModified(false);

  return true;
}

void DesignerDoc::deleteContents()
{
  /////////////////////////////////////////////////
  // TODO: Add implementation to delete the document contents
  /////////////////////////////////////////////////

}

void DesignerDoc::selectWrite( int id){
  write = view->write(id);
}

void DesignerDoc::changeFileName(QString newname){
  if ( newname.isEmpty() ) return;
  QString oldname = write ->fileName();
  if ( newname.left(5) == "file:") newname.remove(0,5);
  if ( oldname.left(5) == "file:") oldname.remove(0,5);

  int *id = new int;
  *id = view->fileTabBar->currentTab();

  int pos;
  QString shortname( newname );
  while ( ( pos = shortname.find('/')) != -1)
    shortname.remove( 0,pos+1 );

  view->fileTabBar->setTabLabel( *id, shortname.data() );

  fileList->remove( oldname.data() );
  fileList->insert( newname.data(), id );

  fileNameList->remove( fileNameList->find(oldname) );
  fileNameList->insert( 0, newname );
}

/** set modified flag */
void DesignerDoc::setModified(bool _m=true) {
  write->setModified(_m);
}

/** get modified flag */
bool DesignerDoc::isModified(){
  return write ? write->isModified() : false;
}

/** save all files */
bool DesignerDoc::saveAll(bool type=true){
  bool flagsave = true;
  int tek_id = view->fileTabBar->currentTab();

  QList<QTab> *list = view->fileTabBar->tabsList;

  for ( list->first(); list->current(); list->next() ) {
    int id = list->current()->id;

    view->fileTabBar->setCurrentTab( id);

    KWrite *twrite = (KWrite *)(view->writeStack->widget( id));
    if ( twrite->isModified() )
      if ( type ) {
        twrite->save();
        changeFileName( QString(twrite->fileName()) );
        if ( twrite->isModified() ) flagsave = false;
      }
      else if ( !twrite->canDiscard() ) return false;
  }

  view->fileTabBar->setCurrentTab( tek_id);
  view->fileTabBar->repaint();
  return flagsave;
}

bool DesignerDoc::isModifiedAll(){
  QList<QTab> *list = view->fileTabBar->tabsList;

  bool modified = false;

  for ( list->first(); list->current(); list->next() ) {
    int id = list->current()->id;
    KWrite *twrite = (KWrite *)(view->writeStack->widget( id));
    if ( twrite->isModified() ) modified = true;
  }

  return modified;
}

/** create new write classa */
KWrite* DesignerDoc::newWrite(QWidget *parent)
{
  HlManager *hl = new HlManager();
  KWriteDoc *writeDoc = new KWriteDoc( hl);

 	write = new KWrite( writeDoc, parent);
 	write -> readConfig( kapp->getConfig() );
	int i = hl->nameFind( "HTML" );
	Highlight* phlHTML = hl->getHl( i );
	phlHTML->use();
	i = hl->nameFind( "PHP" );
	Highlight* phl = hl->getHl( i );
	phl->setSecondary( phlHTML );
	phl->setMode( "HTML" );
 	write -> setHl(i);
	connect( write, SIGNAL( toggleBreakpoint( const QString&, int ) ), SIGNAL( notifyBPChange( const QString&, int ) ) );

/* TODO 	
 	while ( ( i= app->bookmarksMenu->count()) > 3)
    app->bookmarksMenu->removeItemAt(i-1);
  write -> installBMPopup( app->bookmarksMenu );
 	
 	connect( write, SIGNAL( newUndo()), app, SLOT( slotNewUndo()));
 	connect( write, SIGNAL( newStatus()), app, SLOT( slotNewStatus()));
 	connect( write, SIGNAL( markStatus(KWriteView *, bool)), app, SLOT( slotMarkStatus(KWriteView *, bool)));
 	connect( write, SIGNAL( clipboardStatus(KWriteView *, bool)), app, SLOT( slotClipboardStatus(KWriteView *, bool)));
*/ 	

  DesignerApp* app = (DesignerApp*)this->parent();

 	writeRBMenu = new QPopupMenu();
  writeRBMenu->insertItem(i18n("Cu&t"),   app, SLOT(slotEditCut()),   0, ID_EDIT_CUT);
  writeRBMenu->insertItem(i18n("&Copy"),  app, SLOT(slotEditCopy()),  0, ID_EDIT_COPY);
  writeRBMenu->insertItem(i18n("&Paste"), app, SLOT(slotEditPaste()), 0, ID_EDIT_PASTE);
  writeRBMenu->insertSeparator();
  writeRBMenu->insertItem(i18n("&Edit current tag"), this, SLOT(slotEditCurrentTag()),0, ID_EDIT_CURRENT_TAG);
	connect( writeRBMenu, SIGNAL( aboutToShow() ), this, SLOT(slotOnShowRBMenu()) );

 	write->installRBPopup(writeRBMenu);

 	return write;
}

void	DesignerDoc::setStepLine( int iLine )
{
	write->clearStepLine();
	if ( -1 != iLine ){
		write->setStepLine( iLine - 1 );
	}
}

bool isInWord( char c )
{
	return ( 'A' <= c && c <= 'Z' ) || ( 'a' <= c && c <= 'z' );
}

QString DesignerDoc::GetCurrentTag()
{
	TextLine* line = write->doc()->textLine( write->currentLine() );
	int col = write->currentColumn();
	if ( line->length() < col ){
		return "";
	}
	const char*	str = line->getText();
	int iStart = col;
	while ( 0 < iStart && isInWord( str[ iStart - 1 ] ) ){
		--iStart;
	}
	int iEnd = col;
	while ( iEnd < line->length() && isInWord( str[ iEnd ] ) ){
		++iEnd;
	}
	if ( 0 < iStart && '<' == str[ iStart - 1 ] ){
		return QString( str + iStart, iEnd - iStart + 1 );
	}
	else{
		return "";
	}
}

void	DesignerDoc::slotOnShowRBMenu()
{
	QString str = GetCurrentTag();
	writeRBMenu->setItemEnabled( ID_EDIT_CURRENT_TAG, !str.isEmpty() );
}

void	DesignerDoc::slotEditCurrentTag()
{
}

QString DesignerDoc::getBreakpoints( const char* pcszFileName )
{
	QString rc;
	if ( !fileList->find( pcszFileName ) ){
		return "";
	}
	int id = *fileList->find( pcszFileName );
	if ( 0 != id ){
		KWrite* pwrite = view->write( id );
		KWriteDoc* wd = pwrite->doc();
		for ( int i = 0 ; i < wd->getTextLineCount() ; ++i ){
			TextLine* tl = wd->textLine( i );
			if ( 0 != tl->getBPId() ){
				QString strBP;
				rc += strBP.setNum( i + 1 );
				rc += " ";
			}
		}
	}
	return rc;
}

void DesignerDoc::setWriteMode( int i )
{
	Highlight* phl = write->getHighlight();
	QStrList lststr;
	phl->appendModes( lststr );	
	lststr.first();
	const char* pcsz = lststr.current();
	while ( i-- ){
		lststr.next();
		pcsz = lststr.current();
	}
	if ( lststr.current() ){
		phl->setMode( lststr.current() );
		write->setHl( write->getHl() );
	}
}

void DesignerDoc::updateViewMenu( QPopupMenu* viewMenu )
{
  // Clear the highlight options from the view menu
  while ( ID_VIEW_HIGHLIGHT <= viewMenu->idAt( viewMenu->count() - 1 )  ){
  	viewMenu->removeItemAt( viewMenu->count() - 1 );
  }
  // And install the new ones...
  Highlight* phl = write->getHighlight();
  QStrList lststr;
  phl->appendModes( lststr );	
  int i = 0;
  lststr.first();
  while ( lststr.current() ){
  	viewMenu->insertItem( lststr.current(), ID_VIEW_HIGHLIGHT + i );
  	++i;
  	lststr.next();
  }
}