/***************************************************************************
                          designerview.cpp  -  description
                             -------------------
    begin                : Wed May 24 16:16:04 EDT 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qprinter.h>
#include <qpainter.h>

// application specific includes
#include "designerview.h"
#include "designerdoc.h"
#include "designer.h"

#include "./kwrite/kwdoc.h"
#include "./kwrite/highlight.h"

HlManager hlManager; //highlight manager

DesignerView::DesignerView(QWidget *parent, const char *name) :
	QWidget( parent, name )
{
	topLayout = new QVBoxLayout( this );
  writeStack = new QWidgetStack(this);

  fileTabBar  = new FileTabBar( this);
  fileTabBar  ->setFocusPolicy(QWidget::NoFocus);

  topLayout ->addWidget( writeStack,10);
  topLayout ->addWidget( fileTabBar);

  KWrite *write = getDocument()->newWrite(this);

  addWrite(write,"Untitled.php");

	connect( fileTabBar,SIGNAL(selected(int)), this, SLOT( selectWrite(int)));
	
	topLayout->activate();
}

DesignerView::~DesignerView()
{
}

DesignerDoc *DesignerView::getDocument() const
{
  return gpMainFrame->getDocument();
}

void DesignerView::print(QPrinter *pPrinter)
{
  QPainter printpainter;
  printpainter.begin(pPrinter);
	
  // TODO: add your printing code here

  printpainter.end();
}

/** Add new kwrite class to writeStack and return id in stack */
int DesignerView::addWrite( KWrite *w , char *label )
{
  connect( w, SIGNAL(newCurPos()), this, SLOT(slotNewCurPos()));
	int id = fileTabBar->addTab( label );
	writeStack->addWidget( w, id );
	writeStack->raiseWidget( id );
		
	fileTabBar->repaint();
	writeStack->show();
	
		
	return id;
}

/** return current KWrite class */
KWrite * DesignerView::write()
{
  return ( KWrite* )writeStack->visibleWidget();
}

/** set new visible write */
void DesignerView::selectWrite(int id)
{
  emit writeSelected(id); // Notify the doc that the current write has changed

	if ( -1 != id ){
	  writeStack->raiseWidget(id);
	  QString caption=kapp->getCaption();	
		gpMainFrame->setCaption( caption + ": " + gpMainFrame->getDocument()->getTitle());
	}

  slotNewCurPos();
}

/** remove KWrite class from stack, return id of new KWrite */
KWrite* DesignerView::removeWrite()
{
  KWrite *w = write();
  int id = writeStack->id(w);
  fileTabBar->removeTab(id);
  writeStack->removeWidget( w );
  writeStack->raiseWidget( fileTabBar->currentTab() );

  return write();
}

/** return write */
KWrite * DesignerView::write(int id)
{
  return (KWrite*)writeStack->widget(id);
}

/**  */
void DesignerView::slotNewCurPos()
{
   emit newCurPos();
}
