/***************************************************************************
                          phpdbgr.cpp  -  description
                             -------------------
    begin                : Mon Jun 5 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <assert.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <stdlib.h>

#include <kmsgbox.h>

#include "designerdoc.h"
#include "phpdbgr.h"

CPHPDebugger::CPHPDebugger( DesignerApp* papp )
	: CDebugger( papp )
{
	m_cbBuf = 0;
	m_psnSvr = 0;
	m_sockSvr = -1;
	m_psnSession = 0;
	m_sockSession = -1;
}

CPHPDebugger::~CPHPDebugger()
{
}

int	CPHPDebugger::curLine()
{
	return m_line;
}

QString CPHPDebugger::EscapeExpr( const char* pcsz )
{
	QString rc;
	rc.resize( 2 * strlen( pcsz ) + 1 );
	int i = 0;
  while ( '\0' != *pcsz ){
		switch ( *pcsz ){
		case ' ':
			rc[ i ] = '\\';
			++i;
			rc[ i ] = 's';
			break;
		case '\\':
			rc[ i ] = '\\';
			++i;
			rc[ i ] = '\\';
			break;
		default:
			rc[ i ] = *pcsz;
		}
		++i;
		++pcsz;
	}
	rc[ i ] = '\0';
	rc.resize( i + 1 );
	return rc;
}

QString CPHPDebugger::UnEscapeExpr( const char* pcsz )
{
	QString rc;
	rc.resize( strlen( pcsz ) + 1 );
	int i = 0;
	while ( '\0' != *pcsz ){
		if ( '\\' != *pcsz ){
			rc[ i ] = *pcsz;
		}
		else{
			++pcsz;
			switch( *pcsz ){
			case '\\':
				rc[ i ] = '\\';
				break;
			case 's':
				rc[ i ] = ' ';
				break;
			default:
				assert( FALSE );
			}
		}
		++i;
		++pcsz;
	}
	rc[ i ] = '\0';
	rc.resize( i + 1 );
	return rc;
}

void	CPHPDebugger::OnDataReady( void )
{
	int cbRead = recv( m_sockSession, m_acBuffer + m_cbBuf, 1024 - m_cbBuf, 0 );
	if ( 0 < cbRead ){
		m_cbBuf += cbRead;
		UpdateState();
	}
	else if ( 0 == cbRead  && 0 == errno ){
		m_state = LISTENING;
		close( m_sockSession );
		m_sockSession = -1;
		delete m_psnSession;
		m_psnSession = 0;
		UpdateState();
	}
}

void  CPHPDebugger::OnProcessExited( KProcess* proc )
{
	//	Double check to terminate the debug session if debuggee never managed to connect to us
	if ( INACTIVE != m_state ){
		m_state = INACTIVE;
	}
	if ( !proc->normalExit() || 0 != proc->exitStatus() ){
		KMsgBox::message( m_papp, kapp->getCaption(), i18n( "Unable to launch php" ), KMsgBox::EXCLAMATION );
	}
	delete proc;
}

void	CPHPDebugger::OnConnectReady( void )
{
	//	TODO - need to reject the connection gracefully or something
	if ( -1 != m_sockSession ){
		return;
	}
	struct sockaddr_in	cliaddr;
	socklen_t clilen = sizeof( cliaddr );
	m_sockSession = accept( m_sockSvr, (struct sockaddr*)&cliaddr, &clilen );
	if ( 0 <= m_sockSession ){
		int fNoBlock = 1;
		ioctl(  m_sockSession, FIONBIO, &fNoBlock );
		m_psnSession = new QSocketNotifier( m_sockSession, QSocketNotifier::Read );
		QObject::connect( m_psnSession, SIGNAL( activated( int ) ), this, SLOT( OnDataReady() ) );
	}
	m_line = 0;
	m_state = BUSY;	//	Connected, but don't send any commands yet
}

int	CPHPDebugger::fIsActive()
{
	return ( INACTIVE != m_state );
}

void	CPHPDebugger::Activate()
{
	m_papp->ClearDbgWnds();
	
	m_sockSvr = socket( AF_INET, SOCK_STREAM, 0 );
	int fNoBlock = 1;
	ioctl(  m_sockSvr, FIONBIO, &fNoBlock );
	struct sockaddr_in	servaddr;
	bzero( &servaddr, sizeof( servaddr ) );
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons( 6789 );
	servaddr.sin_addr.s_addr = htonl( INADDR_ANY );

	bind( m_sockSvr, (sockaddr*)&servaddr, sizeof( servaddr ) );

	listen( m_sockSvr, SOMAXCONN );
	m_state = LISTENING;

	m_psnSvr = new QSocketNotifier( m_sockSvr, QSocketNotifier::Read );
	QObject::connect( m_psnSvr, SIGNAL( activated(int)), this, SLOT( OnConnectReady() ) );	

	//	TODO	need to add an option for the user to run e.g. a browser here
/*
  KProcess* proc = new KProcess;

	*proc << "/opt/projects/php-4.0.0/php";
	*proc << m_papp->getDocument()->getAbsFilePath();
	connect( proc, SIGNAL( processExited( KProcess* ) ), SLOT( OnProcessExited( KProcess* ) ) );
	if ( !proc->start() ){
		delete proc;
		KMsgBox::message( m_papp, kapp->getCaption(), i18n( "Unable to launch php" ), KMsgBox::EXCLAMATION );
	}
*/
}

void	CPHPDebugger::Go()
{
	if ( READY == m_state ){
		SendLine( "GO" );
		m_state = BUSY;
	}
}

void	CPHPDebugger::StepInto()
{
	if ( READY == m_state ){
		SendLine( "STEP" );
		m_state = BUSY;
	}
}

void	CPHPDebugger::StepOver()
{
	if ( READY == m_state ){
		SendLine( "OVER" );
		m_state = BUSY;
	}
}

void	CPHPDebugger::Stop()
{
	switch ( m_state ){
	case READY:
	case BUSY:
		SendLine( "STOP" );
		break;
	case LISTENING:
		m_state = INACTIVE;
		delete m_psnSvr;
		m_psnSvr = 0;
		close( m_sockSvr );
		m_sockSvr = 0;
	default:
		//	No action required
		;
	}		
}

void	CPHPDebugger::SendLine( const char* pcsz )
{
	QString	str( pcsz );
	str += "\n";
	send( m_sockSession, str, str.length(), 0 );
	m_papp->AppendTranscript( "> " );
	m_papp->AppendTranscript( pcsz );
	m_papp->AppendTranscript( "\r\n" );
}

//	Read until we hit a \n and return excluding the \n
QString	CPHPDebugger::ReadLine()
{
	//	Abort if no complete line stored
	char* pcEOL = (char*)memchr( m_acBuffer, '\n', m_cbBuf );
	if ( NULL == pcEOL ){
		return "";
	}

	//	'\n's within the line are escaped - unescape here before returning
	char szBuffer[ 1024 ];
	int iIn = 0;
	int iOut = 0;
	while ( iIn < pcEOL - m_acBuffer ){
		if ( '\\' != m_acBuffer[ iIn ] ){
			szBuffer[ iOut ] = m_acBuffer[ iIn ];
		}
		else{
			++iIn;
			switch ( m_acBuffer[ iIn ] ){
			case '\\':
				szBuffer[ iOut ] = '\\';
				break;
			case 'n':
				szBuffer[ iOut ] = '\n';
				break;
			default:
				szBuffer[ iOut ] = m_acBuffer[ iIn ];
				assert( FALSE );
			}
		}
		++iIn;
		++iOut;
	}

	//	Save the string to return
	QString rc( szBuffer, iOut + 1 );

	//	and move up the buffer
	m_cbBuf -= ( pcEOL + 1 - m_acBuffer );
	memmove( m_acBuffer, pcEOL + 1, m_cbBuf );

	m_papp->AppendTranscript( rc );
	m_papp->AppendTranscript( "\r\n" );

	return rc;
}

void	CPHPDebugger::OnWatchListChanged()
{
	if ( READY == m_state ){
		QString str = m_papp->PrepareWatchList();
		str = "SETW " + str;
		SendLine( str );
	}
}

void	CPHPDebugger::AddVar( const char* pcszLine, int i )
{
	char* pcSpace = strchr( pcszLine, ' ' );
	QString strParent = UnEscapeExpr( QString( pcszLine, pcSpace - pcszLine + 1 ) );
	char* pcSpace2 = strchr( pcSpace + 1, ' ' );
	QString strName = UnEscapeExpr( QString( pcSpace + 1, pcSpace2 - pcSpace ) );
	QString strValue( pcSpace2 + 1 );
	m_papp->AddVariable( i, strParent, strName, strValue );
}

void	CPHPDebugger::UpdateState()
{
	int fLineModified = FALSE;
	int	fFileModified = FALSE;

	QString str = ReadLine();
	while ( !str.isEmpty() ){
		if ( str == "REDY" ){
			m_state = READY;
		}
		else if ( str == "BEGW" ){
			QString strCmd = m_papp->PrepareWatchList();
			strCmd = "SETW " + strCmd;
			SendLine( strCmd );
		}
		else if ( strncmp( str, "FILE ", 5 ) == 0 ){
			char* pcSpace = strchr( (const char*)str + 5, ' ' );
			m_strFileId = QString( (const char*)str + 5, pcSpace - str - 5 + 1 );
			m_strFile = QString( pcSpace + 1 );
			fLineModified = TRUE;
			fFileModified = TRUE;
		}
		else if ( strncmp( str, "STMT ", 5 ) == 0 ){
			m_line = atoi( (const char*)str + 5 );
			fLineModified = TRUE;
		}
		else if ( str == "BEGA" ){
			m_papp->PrepareVariableList();
		}
		else if ( strncmp( str, "AUTO ", 5 ) == 0 ){
			AddVar( (const char*)str + 5, 0 );
		}
		else if ( strncmp( str, "LOCL ", 5 ) == 0 ){
			AddVar( (const char*)str + 5, 1 );
		}
		else if ( strncmp( str, "THIS ", 5 ) == 0 ){
			AddVar( (const char*)str + 5, 2 );
		}
		else if ( str == "ENDA" ){
			m_papp->PurgeVariableList();
		}
		else if ( strncmp( str, "WTCH ", 5 ) == 0 ){
			char* pcSpace = strchr( (const char*)str + 5, ' ' );
			QString strParent = UnEscapeExpr( QString( (const char*)str + 5, pcSpace - str - 5 + 1 ) );
			char* pcSpace2 = strchr( pcSpace + 1, ' ' );
			QString strName = UnEscapeExpr( QString( pcSpace + 1, pcSpace2 - pcSpace ) );
			QString strValue( pcSpace2 + 1 );
			m_papp->AddWatch( strParent, strName, strValue );
		}
		else if ( str == "ENDW" ){
			m_papp->PurgeWatchList();
		}
		else if ( strncmp( str, "STDO ", 5 ) == 0 ){
			m_papp->AppendStdOut( (const char*)str + 5 );
		}
		else if ( strncmp( str, "STDE ", 5 ) == 0 ){
			m_papp->AppendStdErr( (const char*)str + 5 );
		}
		else if ( strncmp( str, "DASM ", 5 ) == 0 ){
			//TODO	m_papp->...
		}
		else if ( strncmp( str, "HELO ", 5 ) == 0 ){
			m_strServer = QString( (const char*)str + 5 );
			if ( m_strServer == "(null)" ){
				m_strServer = "";
			}
		}
		str = ReadLine();
	}
	if ( fFileModified ){
		SendBreakpoints();
	}
	if ( m_state == READY ){
		m_papp->setStepLine( m_strServer, m_strFile, m_line );
	}
}

void	CPHPDebugger::SendBreakpoints()
{
	QString str =	m_papp->getBreakpoints( m_strServer, m_strFile );
	SendLine( "SETB " + m_strFileId + " " + str );
}

//	NB pcszFile is the local path, not the server path
void CPHPDebugger::slotOnBPChange( const QString& rcstrFile )
{
  QString strFile( rcstrFile );
  if ( strFile.left( 5 ) == "file:" ){
		strFile.remove( 0, 5 );
	}

	if ( m_papp->mapFileName( m_strServer, m_strFile) == strFile ){
		if ( READY == m_state || BUSY == m_state ){
			SendBreakpoints();
		}
	}
}

void CPHPDebugger::slotOnWatchListChange()
{
	if ( READY == m_state ){
		QString strCmd = m_papp->PrepareWatchList();
		strCmd = "SETW " + strCmd;
		SendLine( strCmd );
	}		
}