/***************************************************************************
                          watchwnd.cpp  -  description
                             -------------------
    begin                : Mon Jun 5 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <kapp.h>	//	for i18n
#include <assert.h>

#include "watchwnd.h"
#include "clineeditdlg.h"
#include "designer.h"

CWatchListView::CWatchListView( QWidget* parent ) :
	QListView( parent )
{
}

//	The default implementation doesn't emit rightButtonClicked if 0 == i.  That means you'd need to duplicate
//	the menu handlers to detect right clicks on the window but not on an item.  That's silly.
void CWatchListView::mouseReleaseEvent( QMouseEvent * e )
{
	if ( !e ){
		return;
	}

	if ( e->button() == RightButton && viewport()->rect().contains( e->pos() ) && !itemAt( e->pos() ) ){
		emit rightButtonClicked( 0, viewport()->mapToGlobal( e->pos() ), 0 );
	}
	else{
		QListView::mouseReleaseEvent( e );
	}
}

void CWatchListView::resizeEvent( QResizeEvent* pre )
{
	QListView::resizeEvent( pre );
	setColumnWidth( 1, pre->size().width() - columnWidth( 0 ) );
}

CWatchWnd::CWatchWnd( QWidget* parent, const char* name ) :
	CTabCtl( parent, name )
{
	setShape( QTabBar::TriangularBelow );
	for ( int i = 0 ; i < 4 ; ++i ){
		m_wndTab[ i ] = new CWatchListView( this );
		m_wndTab[ i ]->addColumn( "Name", 75 );
		m_wndTab[ i ]->addColumn( "Value" );
		m_wndTab[ i ]->setAllColumnsShowFocus( TRUE );
		connect( m_wndTab[ i ], SIGNAL( rightButtonClicked( QListViewItem*, const QPoint&, int )),
			SLOT( slotRightClick( QListViewItem* ) ) );
	}
	addTab( m_wndTab[ 0 ], i18n( "Watch 1" ) );
	addTab( m_wndTab[ 1 ], i18n( "Watch 2" ) );
	addTab( m_wndTab[ 2 ], i18n( "Watch 3" ) );
	addTab( m_wndTab[ 3 ], i18n( "Watch 4" ) );
	
	m_popup = new QPopupMenu();
	m_popup->insertItem( i18n( "&Add Watch" ), this, SLOT(slotAddWatch()) );
	m_popup->insertItem( i18n( "&Delete Watch" ), this, SLOT(slotRemoveWatch()) );
	m_popup->insertItem( i18n( "&Change Value" ), this, SLOT(slotChangeValue()) );
}

class CVarListItem : public QListViewItem
{
public:
	CVarListItem( const char* pcszText, CVarListItem* pitemParent, CVarListItem* pitemPrev );
	CVarListItem( const char* pcszText, CVarListItem* pitemParent );
	CVarListItem( const char* pcszText, QListView* plv );

	int	m_extra;
};

CVarListItem::CVarListItem( const char* pcszText, CVarListItem* pitemParent, CVarListItem* pitemPrev ) :
	QListViewItem(pitemParent, pitemPrev, pcszText )
{
}

CVarListItem::CVarListItem( const char* pcszText, CVarListItem* pitemParent ) :
	QListViewItem( pitemParent, pcszText )
{
}

CVarListItem::CVarListItem( const char* pcszText, QListView* plv ) :
	QListViewItem( plv, pcszText )
{
}

CVarListItem* CWatchWnd::firstItem( int i )
{
	return (CVarListItem*)m_wndTab[ i ]->firstChild();
}

CVarListItem* CWatchWnd::nextItem( CVarListItem* pitem )
{
	if ( NULL !=  pitem->firstChild() ){
		return (CVarListItem*)pitem->firstChild();
	}
	while ( NULL != pitem ){
		if ( NULL != pitem->nextSibling() ){
			return (CVarListItem*)pitem->nextSibling();
		}
		pitem = (CVarListItem*)pitem->parent();
	}
	return NULL;
}

/*	Need to escape the returned expression as it is passed directly down the wire */
QString CWatchWnd::PrepareWatchList()
{
	QString rc;
	QStrIList	lststrNames;

	for ( int i = 0 ; i < 4 ; ++i ){
		CVarListItem* pitem = firstItem( i );
		while ( NULL != pitem ){
			if ( NULL == pitem->parent() && -1 == lststrNames.find( pitem->text( 0 ) ) ){
				lststrNames.append( pitem->text( 0 ) );
//	TODO				rc += theApp.pdbgr()->EscapeExpr( pitem->text( 0 ) );
				rc += pitem->text( 0 );
				rc += " ";
			}
			pitem->m_extra = 0;
			pitem = nextItem( pitem );
		}
	}
	return rc;
}

void CWatchWnd::AddVar( const char* pcszParent, const char* pcszName, const char* pcszValue )
{
	for ( int i = 0 ; i < 4 ; ++i ){
		CVarListItem* pitem = firstItem( i );
		while ( NULL != pitem ){
			if ( 0 == strcmp( pitem->text( 0 ), pcszParent ) ){
				CVarListItem* pitemParent = pitem;
				CVarListItem* pitemPrev = NULL;
				int	fChildFound = FALSE;
				pitem = (CVarListItem*)pitem->firstChild();
				while ( NULL != pitem ){
					int iComp = strcmp( pitem->text( 0 ), pcszName );
					if ( 0 <= iComp ){
						fChildFound = ( 0 == iComp );
						break;
					}
					pitemPrev = pitem;
					pitem = (CVarListItem*)pitem->nextSibling();
				}
				if ( !fChildFound ){
					if ( NULL == pitemPrev ){
						pitem = new CVarListItem( pcszName, pitemParent );
					}
					else{
						pitem = new CVarListItem( pcszName, pitemParent, pitemPrev );
					}
				}
				pitem->setText( 1, pcszValue );
				pitem->m_extra = 1;			
			}
			else if ( 0 == strcmp( pitem->text( 0 ), pcszName ) && NULL == pitem->parent() &&
				0 == strcmp( pcszParent, "-" ) ){
				pitem->setText( 1, pcszValue );
				pitem->m_extra = 1;
			}
			pitem = nextItem( pitem );
		}
	}
}

void	CWatchWnd::Purge()
{
  for ( int i = 0 ; i < 4 ; ++i ){
		CVarListItem* pitem = firstItem( i );
		while ( NULL != pitem ){
			if ( 0 == pitem->m_extra ){
				CVarListItem* pitemToGo = pitem;
				pitem = nextItem( pitem );
				delete pitemToGo;
			}
			else{
				pitem = nextItem( pitem );
			}
		}
	}
}

void CWatchWnd::slotRightClick( QListViewItem* pli )
{
	m_popup->setItemEnabled( m_popup->idAt( 1 ), NULL != pli );
	m_popup->setItemEnabled( m_popup->idAt( 2 ), NULL != pli );

	m_pliCurrent = pli;
	m_popup->exec(QCursor::pos());
	m_pliCurrent = NULL;
}

void CWatchWnd::slotAddWatch()
{
	CLineEditDlg dlg( i18n( "Enter the expression to be watched" ), "", gpMainFrame, 0 );
	if ( QDialog::Accepted == dlg.exec() ){
		new CVarListItem( dlg.getText(), m_wndTab[ getCurrentTab() ] );
		emit notifyWatchListChanged();
	}
}

void CWatchWnd::slotRemoveWatch()
{
	if ( NULL != m_pliCurrent ){
		delete m_pliCurrent;
	}
}

void CWatchWnd::slotChangeValue()
{
	if ( NULL != m_pliCurrent ){
	}
}