/***************************************************************************
                          dbgmon.c  -  description
                             -------------------
    begin                : Wed May 31 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <assert.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/time.h>
#include <stdlib.h>

#include <zend_extensions.h>
#include <zend_variables.h>
#include <zend_hash.h>
#include <php.h>
#include <php_globals.h>
#include <SAPI.h>

#include "dbgmon.h"

//	Create and export a zend_extension_version_info
#define ZEND_EXT_API PHP_DBGR_API
ZEND_EXTENSION();

//  Create and export a zend_extension
PHP_DBGR_API	zend_extension	zend_extension_entry = {
	"Nexidion Debug Monitor",	    //	char *name;
	"0.9.1",											//	char *version;
	"Mark_Dixon@teamstudio.com",	//  char *author;
	"http://www.nexidion.org",   	//  char *URL;
	"Copyright Ives Development Inc",	//	char *copyright;

	NULL,													//	int (*startup)(zend_extension *extension);
	NULL,													//	void (*shutdown)(zend_extension *extension);
	onActivate,										//	void (*activate)();
	onDeactivate,									//	void (*deactivate)();

	NULL,													//	void (*message_handler)(int message, void *arg);
	NULL,													//	void (*op_array_handler)(zend_op_array *op_array);
	
	onStatement,									//	void (*statement_handler)(zend_op_array *op_array);
	onFCallBegin,									//	void (*fcall_begin_handler)(zend_op_array *op_array);
	onFCallEnd,										//	void (*fcall_end_handler)(zend_op_array *op_array);

	NULL,													//	void (*op_array_ctor)(void **resource);
	NULL,													//	void (*op_array_dtor)(void **resource);

	STANDARD_ZEND_EXTENSION_PROPERTIES
};

void	TeeStdOut();
void	TeeStdErr();
void	myconnect();

static	int					mhsocket;	//	TODO: These need to be thread-safe...
static	zend_bool		fIsConnected;
static	int			mstdoutSave;	//	Save our original stdout
static	int			mstdoutNew;
static	int			mistdout;
static	int			mstderrSave;
static	int			mstderrNew;
static	int			mistderr;
static	zend_uint		maiBP[20];
static	zend_uint		mcBP;
static	int			mop;
static	char		macBuffer[ 1024 ];
static	int			mcbBuf;
static	int			miStepOverDepth;
static	int			mcFDepth;
static	zend_bool	mfWantRequest;
static	char		mszDebugIP[ 80 ];
static	const char*	mpcszFileLast = "";

#if ZEND_EXTENSION_API_NO == 20000521
//	PHP 4.0.0 version
void    (*old_zend_error)( int type, const char* format, ... );
#elif ZEND_EXTENSION_API_NO == 20000622 || ZEND_EXTENSION_API_NO == 20000815
//	PHP 4.0.1pl2 and PHP 4.0.2 version
void 		(*old_zend_error_cb)(int type, const char *error_filename, const uint error_lineno, const char *format, va_list args);
#else
#error Unsupported PHP version.  Please contact Mark_Dixon@teamstudio.com.
#endif


#define DBGR_STEP	1
#define	DBGR_GO		2

//	Search the config file to see if this request should be debugged.
zend_bool	fWantRequest()
{
	//	Open the config file
	char* pszCfgFile;
	FILE* pfile;
	char  szLine[ 2048 ];
	zend_bool	rc = 0;
	char* pszActualRemote;
	char* pszActualPath;

	pszActualRemote = getenv( "REMOTE_ADDR" );
	pszActualPath = getenv( "PATH_TRANSLATED" );

	//	If REMOTE_ADDR and/or PATH_TRANSLATED don't exist then assume we're command line=>debug locally
	if ( !pszActualRemote || !pszActualPath ){
		strcpy( mszDebugIP, "127.0.0.1" );
		return 1;
	}

	cfg_get_string( "nexidion_config", &pszCfgFile );

	if ( !pszCfgFile || '\0' == pszCfgFile[ 0 ] ){
		return rc;
	}
	pfile = fopen( pszCfgFile, "r" );
	if ( NULL == pfile ){
		return rc;
	}
	
	//	Read until we find a match
	while ( NULL != fgets( szLine, 2048, pfile ) ){
		char* pszRemoteIP = strtok( szLine, "," );
		char* pszPath = strtok( NULL, "," );
		char* pszDebugIP = strtok( NULL, "," );
		char* pszActive = strtok( NULL, "," );
		if ( !pszActive ){ //	Invalid line
			continue;
		}
		if ( ( '\0' == pszRemoteIP[ 0 ] ||  0 == strcmp( pszRemoteIP, pszActualRemote ) ) &&
			( '\0' == pszPath || 0 == strncmp( pszPath, pszActualPath, strlen( pszPath ) ) ) &&
			( 0 == strcmp( pszActive, "Yes" ) ) ){
			strncpy( mszDebugIP, pszDebugIP, 79 );
			rc = 1;
			break;
		}
	}
	fclose( pfile );
	return rc;
}

void	SendData( const char* pcsz )
{
	//	alloc a buffer big enough for the escaped string
	char*	pszBuffer = malloc( 2 * strlen( pcsz ) + 2 );
	int iIn = 0;
	int iOut = 0;

	while ( pcsz[ iIn ] ){
		switch( pcsz[ iIn ] ){
		case '\n':
			pszBuffer[ iOut++ ] = '\\';
			pszBuffer[ iOut ] = 'n';
			break;
		case '\\':
			pszBuffer[ iOut++ ] = '\\';
			pszBuffer[ iOut ] = '\\';
			break;
		default:
			pszBuffer[ iOut ] = pcsz[ iIn ];
		}
		++iOut;
		++iIn;
	}
	pszBuffer[ iOut ] = '\n';
	send( mhsocket, pszBuffer, iOut + 1, 0 );
	free( pszBuffer );
}

#if ZEND_EXTENSION_API_NO == 20000521
PHP_DBGR_API	void	MyZendError( int type, const char* format, ... )
{
	va_list args;
	char	buffer[ 1024 ];
	char* error_filename = NULL;
	uint	error_lineno = 0;
	CLS_FETCH();
	ELS_FETCH();

	va_start( args, format );
	vsnprintf( buffer, sizeof( buffer ) - 1, format, args );
	va_end( args );
	buffer[ sizeof( buffer ) - 1 ] = 0;

	/* If we're an internal call, usually from zend_eval_string, then don't send the error */
	if ( EG(no_extensions) == 1 ){
		old_zend_error( type, buffer );
		return;
	}

	/* If we're executing, treat it as a breakpoint on the line */
	if ( zend_is_executing() ){
		onStatement( EG( active_op_array ) );
	}
	else if ( zend_is_compiling() ){
		char szFileFull[ PATH_MAX ];
		char szBuffer[ 1024 ];

		error_filename = zend_get_compiled_filename( CLS_C );
		error_lineno = zend_get_compiled_lineno( CLS_C );
		if ( NULL != error_filename ){
			if ( fIsConnected ){
				myconnect();
			}
		}
		mpcszFileLast = error_filename;
		mcBP = 0;
		if ( NULL == realpath( error_filename, szFileFull ) ){
			strcpy( szFileFull, error_filename );
		}
		sprintf( szBuffer, "FILE %x %s", (uint)error_filename, szFileFull );

		SendData( szBuffer );
		sprintf( szBuffer, "STMT %u", error_lineno );
		SendData( szBuffer );
		SendData( "REDY" );
	}

	/* Now pass the error on to the old error handler - we do our processing first in case
		 old_zend_error bails out in which case we'd never get called
	*/
	old_zend_error( type, buffer );
}
#elif ZEND_EXTENSION_API_NO == 20000622 || ZEND_EXTENSION_API_NO == 20000815
PHP_DBGR_API	void	MyZendErrorCb( int type, const char *error_filename, const uint error_lineno, const char *format, va_list args )
{
	CLS_FETCH();
	ELS_FETCH();

	/* If we're an internal call, usually from zend_eval_string, then don't send the error */
	if ( EG(no_extensions) == 1 ){
		old_zend_error_cb( type, error_filename, error_lineno, format, args );
		return;
	}

	/* If we're executing, treat it as a breakpoint on the line */
	if ( zend_is_executing() ){
		onStatement( EG( active_op_array ) );
	}
	else if ( zend_is_compiling() ){
		char szFileFull[ PATH_MAX ];
		char szBuffer[ 1024 ];

		if ( NULL != error_filename ){
			if ( fIsConnected ){
				myconnect();
			}
		}
		mpcszFileLast = error_filename;
		mcBP = 0;
		if ( NULL == realpath( error_filename, szFileFull ) ){
			strcpy( szFileFull, error_filename );
		}
		sprintf( szBuffer, "FILE %x %s", (uint)error_filename, szFileFull );

		SendData( szBuffer );
		sprintf( szBuffer, "STMT %u", error_lineno );
		SendData( szBuffer );
		SendData( "REDY" );
	}

	/* Now pass the error on to the old error handler - we do our processing first in case
		 old_zend_error_cb bails out in which case we'd never get called
	*/
	old_zend_error_cb( type, error_filename, error_lineno, format, args );
}
#endif

PHP_DBGR_API	void	onActivate()
{
	char szFile[ 255 ];

	CLS_FETCH();
	SLS_FETCH();

	fIsConnected = 0;
	mfWantRequest = fWantRequest();

	if ( mfWantRequest ){
		CG( extended_info ) = 1;
	}
	else{                          
		return;
	}

#if ZEND_EXTENSION_API_NO == 20000521
	old_zend_error = zend_error;
	zend_error = MyZendError;
#elif ZEND_EXTENSION_API_NO == 20000622 || ZEND_EXTENSION_API_NO == 20000815
	old_zend_error_cb = zend_error_cb;
	zend_error_cb = MyZendErrorCb;
#endif

	mhsocket = socket( AF_INET, SOCK_STREAM, 0 );

	sprintf( szFile, "/tmp/phpdbg-stdout-%d", getpid() );
	mstdoutSave = dup( 1 );		//	Save current stdout
	mstdoutNew = open( szFile, O_CREAT | O_TRUNC | O_RDWR, S_IREAD | S_IWRITE );	//	Redirect
	dup2( mstdoutNew, 1 );
	sprintf( szFile, "/tmp/phpdbg-stderr-%d", getpid() );
	mstderrSave = dup( 2 );
	mstderrNew = open( szFile,  O_CREAT | O_TRUNC | O_RDWR, S_IREAD | S_IWRITE );	//	Redirect
	dup2( mstderrNew, 2 );
	//	Turn off buffering for stdout/err
	setvbuf( stdout, NULL, _IONBF, 0 );
	setvbuf( stderr, NULL, _IONBF, 0 );
	mistdout = 0;
	mistderr = 0;

	mcBP = 0;
	mop = DBGR_STEP;
	mcbBuf = 0;
	miStepOverDepth = 0;
	mcFDepth = 0;
}

PHP_DBGR_API	void	onDeactivate()
{
	char szFile[ 100 ];

	if ( !mfWantRequest ){
		return;
	}

	if ( !fIsConnected ){
		myconnect();
	}

	TeeStdOut();
	TeeStdErr();

	close( mstdoutNew );
	dup2( mstdoutSave, 1 );
	close( mstdoutSave );
	close( mstderrNew );
	dup2( mstderrSave, 2 );
	close( mstderrSave );

	sprintf( szFile, "/tmp/phpdbg-stdout-%d", getpid() );
	unlink( szFile );
	sprintf( szFile, "/tmp/phpdbg-stderr-%d", getpid() );
	unlink( szFile );

	fIsConnected = 0;
	close( mhsocket );
}

void  SendHelo()
{
	zend_bool fLogSave;
	zend_bool fDispSave;
	zval			val;
	char			szSendBuf[ 255 ];

	PLS_FETCH();
	CLS_FETCH();
	ELS_FETCH();

	fLogSave = PG( log_errors );
	fDispSave = PG( display_errors );
	PG( log_errors ) = 0;
	PG( display_errors ) = 0;
	val.type = 0xFF;
	zend_eval_string( "$SERVER_NAME", &val CLS_CC ELS_CC );
	PG( log_errors ) = fLogSave;
	PG( display_errors ) = fDispSave;
	assert( 0xFF != val.type );
	sprintf( szSendBuf, "HELO %s", val.value.str.val );
	SendData( szSendBuf );
}

void	myconnect()
{
	struct sockaddr_in	servaddr;
	int rc;

	bzero( &servaddr, sizeof( servaddr ) );
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons( 6789 );
	servaddr.sin_addr.s_addr = inet_addr( mszDebugIP );

	rc = connect( mhsocket, (struct sockaddr*)&servaddr, sizeof( servaddr ) );

	if ( -1 == rc ){
		char szBuffer[ 256 ];
		sprintf( szBuffer, "Socket error: %d", errno );
	}
	else{
		fIsConnected = 1;
	}

	SendHelo();
}

void	sprintval( char* szBuffer, int cbBufLen, zval* expr )
{
	zval expr_copy;
	int use_copy;

	zend_make_printable_zval(expr, &expr_copy, &use_copy);
	if (use_copy) {
		expr = &expr_copy;
		strncpy( szBuffer, expr->value.str.val, cbBufLen );
		zval_dtor(expr);
	}
	else{
		szBuffer[0] = '\"';
		strncpy( szBuffer + 1, expr->value.str.val, cbBufLen - 3 );
		szBuffer[ cbBufLen - 2 ] = '\0';
		strcat( szBuffer, "\"" );
	}
	szBuffer[ cbBufLen - 1 ] = '\0';
}

void	SendVar( const char* pszType, char* pcszName, char* pcszParent )
{
	zval	val;
	char	szValBuf[ 256 ];
	char	szBuffer[ 256 ];
	zend_bool	fLogSave;
	zend_bool	fDispSave;

	ELS_FETCH();
	CLS_FETCH();
	PLS_FETCH();


	//	zend_eval_string crashes if the string compile fails because the filename/lineno aren't set
	//	We don't really want compile failures logged anyhow, so we'll turn of logging
	fLogSave = PG(log_errors);
	fDispSave = PG(display_errors);
	PG(log_errors) = 0;
	PG(display_errors) = 0;
	val.type = 0xFF;
	zend_eval_string( pcszName, &val CLS_CC ELS_CC );
	PG(log_errors) = fLogSave;
	PG(display_errors) = fDispSave;
	if ( 0xFF == val.type ){
		strcpy( szValBuf, "Parse error" );
	}
	else{
		sprintval( szValBuf, 256, &val );
	}
	sprintf( szBuffer, "%s %s %s %s", pszType, pcszParent ? pcszParent : "-", pcszName, szValBuf );
	if ( IS_ARRAY == val.type ){
		HashPosition pos;
		int c = zend_hash_num_elements( val.value.ht );
		sprintf( szValBuf, " (%d element%s)", c, c == 1 ? "" : "s" );
		strcat( szBuffer, szValBuf );
		SendData( szBuffer );
		zend_hash_internal_pointer_reset_ex( val.value.ht, &pos );
		while ( NULL != pos ){
			char* pszVal;
			ulong nVal;
			ulong cbVal;
			int rc = zend_hash_get_current_key_ex( val.value.ht, &pszVal, &cbVal, &nVal, &pos );
			switch ( rc ){
			case HASH_KEY_IS_STRING:

				sprintf( szBuffer, "%s[\"%s\"]", pcszName, pszVal );
				pefree( pszVal, val.value.ht->persistent );
				SendVar( pszType, szBuffer, pcszName );
				break;
			case HASH_KEY_IS_LONG:
				sprintf( szBuffer, "%s[%lu]", pcszName, nVal );
				SendVar( pszType, szBuffer, pcszName );
				break;
			}
			zend_hash_move_forward_ex( val.value.ht, &pos );
		}
	}
	else if ( IS_OBJECT == val.type ){
		HashPosition pos;
		sprintf( szValBuf, " (%*s)", val.value.obj.ce->name_length, val.value.obj.ce->name );
		strcat( szBuffer, szValBuf );
		SendData( szBuffer );
		zend_hash_internal_pointer_reset_ex( val.value.obj.properties, &pos );
		while ( NULL != pos ){
			char* pszVal;
			ulong nVal;
			ulong cbVal;
			int rc = zend_hash_get_current_key_ex( val.value.obj.properties, &pszVal, &cbVal, &nVal, &pos );
			switch ( rc ){
			case HASH_KEY_IS_STRING:
				sprintf( szBuffer, "%s->%s", pcszName, pszVal );
				pefree( pszVal, val.value.obj.properties->persistent );
				SendVar( pszType, szBuffer, pcszName );
				break;
			}
			zend_hash_move_forward_ex( val.value.obj.properties, &pos );
		}
	}
	else{
		SendData( szBuffer );
	}
	zval_dtor( &val );
}

void	SendAutoVar( zend_op* opline, zend_op* opstart )
{
	char	szBuffer[ 256 ];
	if ( ( ZEND_FETCH_R == opline->opcode || ZEND_FETCH_W == opline->opcode ||
		ZEND_FETCH_RW == opline->opcode ) ){
		if ( IS_CONST == opline->op1.op_type ){
			sprintf( szBuffer, "$%s", opline->op1.u.constant.value.str.val );
			SendVar( "AUTO", szBuffer, 0 );
		}
	}
	else if ( ZEND_FETCH_OBJ_R == opline->opcode || ZEND_FETCH_OBJ_W == opline->opcode ||
		ZEND_FETCH_OBJ_RW == opline->opcode ){
		if ( IS_CONST == opline->op2.op_type && IS_VAR == opline->op1.op_type ){
			//	Scan back to find the fetch for the object
			zend_op* opScan = opline - 1;
			while ( 1 ){
				if ( ( ZEND_FETCH_R == opScan->opcode || ZEND_FETCH_W == opScan->opcode || ZEND_FETCH_RW == opScan->opcode ) &&
					IS_VAR == opScan->result.op_type && opScan->result.u.var == opline->op1.u.var ){
					if ( IS_CONST == opScan->op1.op_type ){
						sprintf( szBuffer, "$%s->%s", opScan->op1.u.constant.value.str.val, opline->op2.u.constant.value.str.val );
						SendVar( "AUTO", szBuffer, 0 );
					}
					break;
				}
				if ( opScan == opstart ){
					break;
				}
				--opScan;
			}
		}
	}
	else if ( ZEND_FETCH_DIM_R == opline->opcode || ZEND_FETCH_DIM_W == opline->opcode ||
		ZEND_FETCH_DIM_RW == opline->opcode ){
		if ( IS_CONST == opline->op2.op_type && IS_VAR == opline->op1.op_type ){
			//	Scan back to find the fetch for the object
			zend_op* opScan = opline - 1;
			while ( 1 ){
				if ( ( ZEND_FETCH_R == opScan->opcode || ZEND_FETCH_W == opScan->opcode || ZEND_FETCH_RW == opScan->opcode ) &&
					IS_VAR == opScan->result.op_type && opScan->result.u.var == opline->op1.u.var ){
					if ( IS_CONST == opScan->op1.op_type ){
						char szVal[ 100 ];
						sprintval( szVal, 100, &opline->op2.u.constant );
						sprintf( szBuffer, "$%s[%s]", opScan->op1.u.constant.value.str.val, szVal );
						SendVar( "AUTO", szBuffer, 0 );
					}
					break;
				}
				if ( opScan == opstart ){
					break;
				}
				--opScan;
			}
		}
	}
}

void	TeeStdOut()
{
	//	Send any stdout we've accrued, and copy it to the original stdout
	char	szBuffer[ 256 ];
	int istdout = lseek( 1, 0, SEEK_CUR );

	while ( mistdout != istdout ){
		char	szRead[ 240 ];
		unsigned int	cbRead;

		lseek( 1, mistdout, SEEK_SET );
		cbRead = read( 1, szRead, 239 );
		szRead[ cbRead ] = '\0';
		write( mstdoutSave, szRead, cbRead );
		sprintf( szBuffer, "STDO %s", szRead );
		SendData( szBuffer );

		mistdout += cbRead;
	}
}

void	TeeStdErr()
{
	//	Send any stderr we've accrued, and copy it to the original stdout
	char	szBuffer[ 256 ];
	int istderr = lseek( 2, 0, SEEK_CUR );

	while ( mistderr != istderr ){
		char	szRead[ 240 ];
		unsigned int 	cbRead;

		lseek( 2, mistderr, SEEK_SET );
		cbRead = read( 2, szRead, 239 );
		szRead[ cbRead ] = '\0';
		write( mstderrSave, szRead, cbRead );
		sprintf( szBuffer, "STDE %s", szRead );
		SendData( szBuffer );

		mistderr += cbRead;
	}
}

void	SetBreakpoints( const char* pcsz )
{
	int		iLine;
	char*	pszFile;
	mcBP = 0;

	//	Check we're setting breakpoints for the current file!
	if ( 1 == sscanf( pcsz, "%x", (uint*)&pszFile ) && pszFile != mpcszFileLast ){
		return;
	}
	while ( *pcsz != ' ' && *pcsz != '\0' ){
		++pcsz;
	}
	while ( *pcsz == ' ' ){
		++pcsz;
	}


	iLine = atoi( pcsz );
	while ( 0 != iLine ){
		maiBP[ mcBP++ ] = iLine;
		while ( *pcsz != ' ' && *pcsz != '\0' ){
			++pcsz;
		}
		while ( *pcsz == ' ' ){
			++pcsz;
		}
		iLine =atoi( pcsz );
	}
}

//	Read until we hit a \n and return excluding the \n
void	ReadLine( char* pszRet, zend_bool fBlock )
{
	char* pcEOL = (char*)memchr( macBuffer, '\n', mcbBuf );
	char	szBuffer[ 1024 ];
	int iIn = 0;
	int iOut = 0;

	if ( !fIsConnected ){
		strcpy( pszRet, "STEP" );
		return;
	}

	//	Wait if no complete line stored
	while ( NULL == pcEOL ){
		fd_set readfds;
		int cbRead;
		FD_ZERO( &readfds );
		FD_SET( mhsocket, &readfds );
		if ( fBlock ){
			select( mhsocket + 1, &readfds, NULL, NULL, NULL );
			cbRead = recv( mhsocket, macBuffer + mcbBuf, 1023 - mcbBuf, 0 );
		}
		else{
			struct timeval tv = { 0, 0 };
			select( mhsocket + 1, &readfds, NULL, NULL, &tv );
			if ( FD_ISSET( mhsocket, &readfds ) ){
				cbRead = recv( mhsocket, macBuffer + mcbBuf, 1023 - mcbBuf, 0 );
			}
			else{
				cbRead = 0;
			}
		}
		if ( cbRead <= 0 ){
			pszRet[ 0 ] = '\0';
			return;
		}
		else{
			mcbBuf += cbRead;
		}
		pcEOL = (char*)memchr( macBuffer, '\n', mcbBuf );
	}

	//	De-escape the line
	while ( iIn < pcEOL - macBuffer ){
		if ( '\\' != macBuffer[ iIn ] ){
			szBuffer[ iOut ] = macBuffer[ iIn ];
		}
		else{
			++iIn;
			switch ( macBuffer[ iIn ] ){
			case '\\':
				szBuffer[ iOut ] = '\\';
				break;
			case 'n':
				szBuffer[ iOut] = '\n';
				break;
			}
		}
		++iIn;
		++iOut;
	}
	//	Save the line to return
	memcpy( pszRet, szBuffer, iOut );
	pszRet[ iOut ] = '\0';

	//	And move up the buffer
	mcbBuf -= ( pcEOL + 1 - macBuffer );
	memmove( macBuffer, pcEOL + 1, mcbBuf );
}

void  ProcessOOBInput( zend_bool fBlock )
{
	char szBuffer[ 1024 ];
	int fExit = 0;

	while ( !fExit ){
		ReadLine( szBuffer, fBlock );
		if ( 0 == szBuffer[ 0 ] ){
			if ( fBlock ){
				zend_bailout();
			}
			fExit = 1;
		}
		else if ( 0 == strcmp( szBuffer, "STOP" ) ){
			zend_bailout();
			fExit = 1;
		}
		else if ( 0 == strncmp( szBuffer, "SETB ", 5 ) ){
			SetBreakpoints( szBuffer + 5 );
			fExit = 1;
		}
	}
}

void	DumpZnode( char* pszBuffer, znode* pnode )
{
	char	szVarBuf[ 10 ];
	switch( pnode->op_type ){
	case IS_CONST:
		strcat( pszBuffer, " C " );
		sprintval( pszBuffer + strlen( pszBuffer ), 256, &pnode->u.constant );
		break;
	case IS_TMP_VAR:
		strcat( pszBuffer, " T " );
		sprintf( szVarBuf, "%d", pnode->u.var );
		strcat( pszBuffer, szVarBuf );
		break;
	case IS_VAR:
		strcat( pszBuffer, " V " );
		sprintf( szVarBuf, "%d", pnode->u.var );
		strcat( pszBuffer, szVarBuf );
		break;
	default:
		assert( IS_UNUSED == pnode->op_type );
		strcat( pszBuffer, " U " );
	}
}

void	SendWatch( char* szBuffer )
{
	char* pszTok = strtok( szBuffer, " " );
	while ( NULL != pszTok ){
		SendVar( "WTCH", pszTok, 0 );
		pszTok = strtok( NULL, " " );
	}
	SendData( "ENDW" );
}

void	DumpDisasm()
{
/*
	zend_op_array* op_array;
	char* pszFileLast = "";
	zend_uint	i;

	ELS_FETCH();

	SendData( "DASM RESET" );
	op_array = EG( main_op_array );
	for ( i = 0 ; i < op_array->last ; ++i ){
		char	szLine[ 1000 ];
		zend_op* pop = op_array->opcodes + i;
		if ( 0 != strcmp( pszFileLast, pop->filename ) ){
			sprintf( szLine, "DASM FILE %s", pop->filename );
			pszFileLast = pop->filename;
			SendData( szLine );
		}
		sprintf( szLine, "DASM OP %d %d", pop->lineno, pop->opcode );
		DumpZnode( szLine, &pop->result );
		DumpZnode( szLine, &pop->op1 );
		DumpZnode( szLine, &pop->op2 );

		SendData( szLine );
	}
	SendData( "DASM DONE" );
*/
}

void	SendLocals()
{
	HashPosition pos;
	char szBuffer[ 256 ];

	ELS_FETCH();

	//	Don't send anything if we're at the global level - there are too many globals to send
	//	every time
	if ( EG(active_symbol_table) == &EG(symbol_table) ){
		return;
	}

	zend_hash_internal_pointer_reset_ex( EG(active_symbol_table), &pos );
	while ( NULL != pos ){
		char* pszVal;
		ulong nVal;
		ulong cbVal;
		int rc = zend_hash_get_current_key_ex( EG(active_symbol_table), &pszVal, &cbVal, &nVal, &pos );
		switch ( rc ){
		case HASH_KEY_IS_STRING:
			sprintf( szBuffer, "$%s", pszVal );
			SendVar( "LOCL", szBuffer, NULL );
			pefree( pszVal, EG(active_symbol_table)->persistent );
			break;
		}
		zend_hash_move_forward_ex( EG(active_symbol_table), &pos );
	}
}

void	SendThis()
{
	SendVar( "THIS", "$this", NULL );
}

void	SetValue( char* psz )
{
	zend_bool	fLogSave;
	zend_bool	fDispSave;
	char* pszSpace;

	PLS_FETCH();
	CLS_FETCH();
	ELS_FETCH();

	pszSpace = strchr( psz, ' ' );
	*pszSpace = '=';
	fLogSave = PG(log_errors);
	fDispSave = PG(display_errors);
	PG(log_errors) = 0;
	PG(display_errors) = 0;
	strcat( psz, ";" );		//I know, shouldn't write into their buffer...
	zend_eval_string( psz, NULL CLS_CC ELS_CC );
	PG(log_errors) = fLogSave;
	PG(display_errors) = fDispSave;
}

void ProcessInput()
{
	char szBuffer[ 1024 ];
	
	while ( 0 == mop ){
		ReadLine( szBuffer, 1 );
		if ( 0 == strncmp( szBuffer, "SETB ", 5 ) ){
			SetBreakpoints( szBuffer + 5 );
		}
		else if ( 0 == strcmp( szBuffer, "GO" ) ){
			mop = DBGR_GO;
		}
		else if ( 0 == strcmp( szBuffer, "OVER" ) ){
			miStepOverDepth = mcFDepth + 1;
			mop = DBGR_GO;
		}
		else if ( 0 == strcmp( szBuffer, "STEP" ) ){
			mop = DBGR_STEP;
		}
		else if ( 0 == strcmp( szBuffer, "STOP" ) || 0 == strlen( szBuffer ) ){
			zend_bailout();
		}
		else if ( 0 == strcmp( szBuffer, "DASM" ) ){
			DumpDisasm();
		}
		else if ( 0 == strncmp( szBuffer, "SETW ", 5 ) ){
			SendWatch( szBuffer + 5 );
			SendData( "REDY" );
		}
		else if ( 0 == strncmp( szBuffer, "SETV ", 5 ) ){
			SetValue( szBuffer + 5 );
		}
	}
}

PHP_DBGR_API	void	onStatement( zend_op_array *op_array )
{
	zend_op *opline = *( EG( opline_ptr ) );
	char	szBuffer[ 256 ];
  char*	pszCurrentFile;

	ELS_FETCH();
	CLS_FETCH();

	if ( !mfWantRequest ){
		return;
	}

	if ( !fIsConnected ){
		myconnect();
	}

#if ZEND_EXTENSION_API_NO == 20000521 || ZEND_EXTENSION_API_NO == 20000622
//	PHP 4.0.0 and PHP 4.0.1pl2 version
	pszCurrentFile = opline->filename;
#elif ZEND_EXTENSION_API_NO == 20000815
	pszCurrentFile = zend_get_executed_filename();
#endif

	//	Always (even in GO mode) send a FILE statement if we've changed file and wait to see if there are any
	//	breakpoints to be set
	if ( 0 != strcmp( mpcszFileLast, pszCurrentFile ) ){
    char szFileFull[ PATH_MAX ];
		mpcszFileLast = pszCurrentFile;
		mcBP = 0;

		if ( NULL == realpath( pszCurrentFile, szFileFull ) ){
			strcpy( szFileFull, pszCurrentFile );
		}

		sprintf( szBuffer, "FILE %x %s", (unsigned int)pszCurrentFile, szFileFull );

		//	Tell the debugger where we are
		SendData( szBuffer );

		//	And wait for a response
		ProcessOOBInput( 1 );
	}

	if ( DBGR_GO == mop ){
		zend_bool	fBreakHere = 0;
		zend_uint	i;
		if ( mcFDepth < miStepOverDepth ){
			fBreakHere = 1;
		}
		else{
			for ( i = 0 ; i < mcBP ; ++i ){
				if ( maiBP[ i ] == opline->lineno ){
					fBreakHere = 1;
					break;
				}
			}
		}
		if ( !fBreakHere ){
			//	If we're not stopping, do a non-blocking check to see if there are any urgent commands
			ProcessOOBInput( 0 );
			return;
		}
	}
	mop = 0;
	miStepOverDepth = 0;

	TeeStdOut();
	TeeStdErr();

	//	Tell the debugger where we are
	sprintf( szBuffer, "STMT %d", opline->lineno );
	SendData( szBuffer );

	//	Send any required auto data for this line
	SendData( "BEGA" );
	++opline;
	while ( opline->opcode != ZEND_EXT_STMT ){
		SendAutoVar( opline, op_array->opcodes );
		++opline;
	}
	// ... and the previous line
	opline = *( EG( opline_ptr ) );
	if ( opline != op_array->opcodes ){
		--opline;
		while ( opline->opcode != ZEND_EXT_STMT && opline != op_array->opcodes ){
			SendAutoVar( opline, op_array->opcodes );
			--opline;
		}
	}

	//	Send Locals
	SendLocals();

	//	and 'this'
	SendThis();

	//	End of stuff for 'variables' window
	SendData( "ENDA" );

	SendData( "BEGW" );

	//	And await instructions - set breakpoints, send watch data etc
	ProcessInput();
}

PHP_DBGR_API	void	onFCallBegin( zend_op_array* op_array )
{
	if ( !mfWantRequest ){
		return;
	}

	++mcFDepth;
}

PHP_DBGR_API	void	onFCallEnd( zend_op_array* op_array )
{
	if ( !mfWantRequest ){
		return;
	}

	--mcFDepth;
}

