/***************************************************************************
                          profmain.c  -  description
                             -------------------
    begin                : Wed Aug 23 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <assert.h>
#include <string.h>
#include <php.h>
#include <php_globals.h>
#include <zend_execute.h>

#include "phpprof.h"

//	Create and export a zend_extension_version_info
#define ZEND_EXT_API PHPPROF_API
ZEND_EXTENSION();

//  Create and export a zend_extension
PHPPROF_API	zend_extension	zend_extension_entry = {
	"Nexidion PHP Profiler",	    //	char *name;
	"0.1",												//	char *version;
	"Mark_Dixon@teamstudio.com",	//  char *author;
	"http://www.nexidion.org",   	//  char *URL;
	"Copyright Ives Development Inc",	//	char *copyright;

	NULL,													//	int (*startup)(zend_extension *extension);
	NULL,													//	void (*shutdown)(zend_extension *extension);
	onActivate,										//	void (*activate)();
	onDeactivate,									//	void (*deactivate)();

	NULL,													//	void (*message_handler)(int message, void *arg);
	NULL,													//	void (*op_array_handler)(zend_op_array *op_array);
	
	onStatement,									//	void (*statement_handler)(zend_op_array *op_array);
	onFCallBegin,									//	void (*fcall_begin_handler)(zend_op_array *op_array);
	onFCallEnd,										//	void (*fcall_end_handler)(zend_op_array *op_array);

	NULL,													//	void (*op_array_ctor)(void **resource);
	NULL,													//	void (*op_array_dtor)(void **resource);

	STANDARD_ZEND_EXTENSION_PROPERTIES
};

struct timeval	gtvActivate;
struct timeval  gtvDeactivate;
struct timeval  gtvExecuteStart;

static zend_op* 			mopCurrent;
static struct timeval mtvLast;

enum PushType{
PT_NONE = 0,	//	Fn has been pushed
PT_FN,				//	Waiting for a plain function name
PT_METHOD,		//	Waiting for a function name to append to <mszClass>::
PT_THIS				//	Waiting for a function name to append to the class type of '$this'
};
static enum PushType mPushType;
static char 	mszClass[ 256 ];


void UpdateCurrentLine()
{
	if ( NULL == mopCurrent ){
		return;
	}

	BumpLineCountData( mopCurrent, &mtvLast );		
}


void HandlePushType( ELS_D )
{
	char* pszFn;        	char	szBuffer[ 256 ];
	if ( PT_NONE == mPushType ){
		return;
	}
	pszFn = get_active_function_name( ELS_C );
	switch ( mPushType ){
	case PT_FN:
   	strcpy( szBuffer, pszFn );
		break;
	case PT_METHOD:
		strcpy( szBuffer, mszClass );
		strcat( szBuffer, "::" );
		strcat( szBuffer, pszFn );
		break;
	case PT_THIS:
	{
		int fLogSave;
		int fDispSave;
		zval val;
		PLS_FETCH();
		CLS_FETCH();
		fLogSave = PG(log_errors);
		fDispSave = PG(display_errors);
		PG(log_errors) = 0;
		PG(display_errors) = 0;
		val.type = 0xFF;
		zend_eval_string( "$this", &val CLS_CC ELS_CC );
		PG(log_errors) = fLogSave;
		PG(display_errors) = fDispSave;
		if ( IS_OBJECT == val.type ){
			memcpy( szBuffer, val.value.obj.ce->name, val.value.obj.ce->name_length );
			szBuffer[ val.value.obj.ce->name_length ] = '\0';
		}
		else{
			strcpy( szBuffer, "??" );
		}
		strcat( szBuffer, "::" );
		strcat( szBuffer, pszFn );
		zval_dtor( &val );
	}
		break;
	default:
		break;
	}
	PatchFn( szBuffer );
	mPushType = PT_NONE;
}

PHPPROF_API	void	onActivate()
{
	CLS_FETCH();

	gettimeofday( &gtvActivate, NULL );

	CG( extended_info ) = 1;

	InitLineCountData();
	InitFnCountData();
	
	mopCurrent = NULL;
	mPushType = PT_FN;
	PushFn( NULL );
}

PHPPROF_API	void	onDeactivate()
{
	UpdateCurrentLine();

	PopAllFn();

	gettimeofday( &gtvDeactivate, NULL );

	DumpStats();

	TermFnCountData();
	TermLineCountData();
}

PHPPROF_API void  onStatement( zend_op_array* op_array )
{
	ELS_FETCH();

	UpdateCurrentLine();

	if ( NULL == mopCurrent ){
		gettimeofday( &gtvExecuteStart, NULL );		
	}

	HandlePushType( ELS_CC );

	mopCurrent = *( EG( opline_ptr ) );
	gettimeofday( &mtvLast, NULL );
}

//	By the time we leave here, we must
//	a) PushFn with a valid function name and set mPushType to PT_NONE; or
//	b) PushFn( NULL ) and set mPushType to anything other than PT_NONE
PHPPROF_API void  onFCallBegin( zend_op_array* op_array )
{
	char	szBuffer[ 256 ];
	zend_op*	op;
	ELS_FETCH();

	assert( PT_NONE == mPushType );	//	Should have figured it out by now.  If not then we'll have to put
	//	PushType on the Function stack too.

	mPushType = PT_FN;	//	Assume it's an ordinary function call until shown otherwise
	op = *(EG( opline_ptr ) );
	while ( ZEND_EXT_FCALL_END != op->opcode ){
		if ( ZEND_DO_FCALL == op->opcode ){
			if ( IS_CONST == op->op1.op_type ){
				memcpy( szBuffer, op->op1.u.constant.value.str.val, op->op1.u.constant.value.str.len );
				szBuffer[ op->op1.u.constant.value.str.len ] = '\0';
				PushFn( szBuffer );
				mPushType = PT_NONE;
			}
			break;
		}
		else if ( ZEND_INIT_FCALL_BY_NAME == op->opcode ){
  		if ( IS_UNUSED == op->op1.op_type ){
  			if ( IS_STRING == op->op2.op_type ){
  				memcpy( szBuffer, op->op2.u.constant.value.str.val, op->op2.u.constant.value.str.len );
  				szBuffer[ op->op2.u.constant.value.str.len ] = '\0';
  				PushFn( szBuffer );
  				mPushType = PT_NONE;
  			}
  		}
  		else if ( IS_STRING == op->op1.op_type ){
  			memcpy( mszClass, op->op1.u.constant.value.str.val, op->op1.u.constant.value.str.len );
  			mszClass[ op->op1.u.constant.value.str.len ] = '\0';
  			PushFn( NULL );
  			mPushType = PT_METHOD;
  		}
  		else{
  			PushFn( NULL );
  			mPushType = PT_THIS;
  		}
			break;
		}
		++op;
	}
	if ( PT_FN == mPushType ){
		PushFn( NULL );
	}
}

PHPPROF_API void  onFCallEnd( zend_op_array* op_array )
{
	PopFn();
	mPushType = PT_NONE;
}
