/***************************************************************************
                          profstat.c  -  description
                             -------------------
    begin                : Tue Aug 29 2000
    copyright            : (C) Ives Development Inc, 2000
    email                : Mark_Dixon@teamstudio.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>
#include "phpprof.h"

typedef char* StringEntry;

struct LineStats
{
	int							iStringFile;
	int							iLineNo;
	int 						count;
	struct timeval	tvTime;
};

struct FnStats
{
	int							iStringFnName;
	int							count;
	struct timeval  tvTime;
	struct timeval	tvChildTime;
};

struct FnStackEntry
{
	int							iStatTable;
	struct timeval	tvStart;
};

static struct LineStats	maLineStats[ 1000 ];	//	Naive implementation -replace with something more dynamic
static int		cLineStat;
static struct FnStats	maFnStats[ 200 ];
static int		cFnStat;
static struct FnStackEntry maFnStack[ 100 ];
static int		cFnStack;
static StringEntry	maStringTable[ 200 ];
static int 		cString;

//	Utility functions
//	Subtract tv2 from tv1
void SubtractTimeVal( struct timeval* ptv1, struct timeval* ptv2 )
{
	ptv1->tv_sec -= ptv2->tv_sec;
	if ( ptv1->tv_usec < ptv2->tv_usec ){
		--ptv1->tv_sec;
		ptv1->tv_usec += CLOCKS_PER_SEC;
	}
	ptv1->tv_usec -= ptv2->tv_usec;
}

//	Add tv2 to tv1
void AddTimeVal( struct timeval* ptv1, struct timeval* ptv2 )
{
	ptv1->tv_sec += ptv2->tv_sec;
	ptv1->tv_usec += ptv2->tv_usec;
	if ( CLOCKS_PER_SEC <= ptv1->tv_usec ){
		ptv1->tv_usec -= CLOCKS_PER_SEC;
		++ptv1->tv_sec;
	}
}

//  Locate a string in the string table, adding if necessary.
int	FindString( const char* pcsz )
{
	int i ;
	for ( i = 0 ; i < cString ; ++i ){
		if ( 0 == strcmp( pcsz, maStringTable[ i ] ) ){
			break;
		}
	}
	if ( i == cString ){
		maStringTable[ i ] = (char*)malloc( strlen( pcsz ) + 1 );
		strcpy( maStringTable[ i ], pcsz );
		++cString;
	}
	return i;
}

//	Line Stats
void InitLineCountData()
{
	cLineStat = 0;
	cString = 0;
}

void TermLineCountData()
{
	int i;
	for ( i = 0 ; i < cString ; ++i ){
		free( maStringTable[ i ] );
	}
}

void BumpLineCountData( zend_op* op, struct timeval* ptvLast )
{
	int i;
	struct timeval tvNow;
	int iString;

	gettimeofday( &tvNow, NULL );

	iString = FindString( op->filename );
	for ( i = 0 ; i < cLineStat ; ++i ){
		if ( maLineStats[ i ].iStringFile == iString && maLineStats[ i ].iLineNo == op->lineno ){
			break;
		}
	}
	if ( i == cLineStat ){
		assert( cLineStat < 1000 );
		maLineStats[ i ].iStringFile = iString;
		maLineStats[ i ].iLineNo = op->lineno;
		maLineStats[ i ].count = 0;
		maLineStats[ i ].tvTime.tv_sec = 0;
		maLineStats[ i ].tvTime.tv_usec = 0;
		++cLineStat;
	}
	++maLineStats[ i ].count;
	SubtractTimeVal( &tvNow, ptvLast );
	AddTimeVal( &maLineStats[ i ].tvTime, &tvNow );
}

//	Function stats
void	InitFnCountData()
{
	cFnStat = 0;
	cFnStack = 0;
}

void  TermFnCountData()
{
	//	NOP
}

int   GetFnIndex( const char* pcszFn )
{
	int i;
	int iString;
	
	iString = FindString( pcszFn );
	for ( i = 0 ; i < cFnStat ; ++i ){
		if ( maFnStats[ i ].iStringFnName == iString ){
			break;
		}
	}
	if ( cFnStat == i ){
		assert( cFnStat < 200 );
		maFnStats[ i ].iStringFnName = iString;
		maFnStats[ i ].count = 0;
		maFnStats[ i ].tvTime.tv_sec = 0;
		maFnStats[ i ].tvTime.tv_usec = 0;
		maFnStats[ i ].tvChildTime.tv_sec = 0;
		maFnStats[ i ].tvChildTime.tv_usec = 0;
		++cFnStat;
	}
	return i;
}

void	PushFn( const char* pcszFn )
{
	assert( cFnStack < 100 );
	if ( NULL == pcszFn ){
		maFnStack[ cFnStack ].iStatTable = -1;
	}
	else{
		maFnStack[ cFnStack ].iStatTable = GetFnIndex( pcszFn );
	}
	gettimeofday( &maFnStack[ cFnStack ].tvStart, NULL );
	++cFnStack;	
}

void	PatchFn( const char* pcszFn )
{
	assert( 0 < cFnStack );
	assert( -1 == maFnStack[ cFnStack - 1 ].iStatTable );
	maFnStack[ cFnStack - 1 ].iStatTable = GetFnIndex( pcszFn );
}

void	PopFn()
{
	struct timeval tvNow;
  int						 iStatTable;

	if ( 0 == cFnStack ){
		//assert false; - can't pop an empty stack
		return;
	}

	gettimeofday( &tvNow, NULL );

	if ( -1 == maFnStack[ cFnStack - 1 ].iStatTable ){
		PatchFn( "??" );
	}
	iStatTable = maFnStack[ cFnStack - 1 ].iStatTable;	
	++maFnStats[ iStatTable ].count;
	SubtractTimeVal( &tvNow, &maFnStack[ cFnStack - 1 ].tvStart );
	AddTimeVal( &maFnStats[ iStatTable ].tvTime, &tvNow );
	if ( 1 < cFnStack ){
		int iStatTableCaller = maFnStack[ cFnStack - 2 ].iStatTable;
		AddTimeVal( &maFnStats[ iStatTableCaller ].tvChildTime, &tvNow );
	}
	--cFnStack;	
}

void PopAllFn()
{
	while ( 0 < cFnStack ){
		PopFn();
	}
}

//	Final report
void DumpStats()
{
	char szFile[ 256 ];
	FILE* pfile;
	int		i;
	time_t now = time( NULL );
	strftime( szFile, 256, "/tmp/prof-%Y%m%d%H%M%S", localtime( &now ) );
	pfile = fopen( szFile, "w" );
	fprintf( pfile, ";Activate, ExecuteStart, Deactivate\n%ld,%ld,%ld,%ld,%ld,%ld\n",
		gtvActivate.tv_sec, gtvActivate.tv_usec,
		gtvExecuteStart.tv_sec, gtvExecuteStart.tv_usec,
		gtvDeactivate.tv_sec, gtvDeactivate.tv_usec );
	fprintf( pfile, ";Line Stats\n" );
	for ( i = 0 ; i < cLineStat ; ++i ){
		fprintf( pfile, "%s,%d,%d,%ld,%ld\n", maStringTable[ maLineStats[ i ].iStringFile ],
			maLineStats[ i ].iLineNo, maLineStats[ i ].count, maLineStats[ i ].tvTime.tv_sec,
			maLineStats[ i ].tvTime.tv_usec );
		fflush( pfile );
	}
	fprintf( pfile, "\n;Function Stats\n" );
	for ( i = 0 ; i < cFnStat ; ++i ){
		fprintf( pfile, "%s,%d,%ld,%ld,%ld,%ld\n", maStringTable[ maFnStats[ i ].iStringFnName ], maFnStats[ i ].count,
			maFnStats[ i ].tvTime.tv_sec, maFnStats[ i ].tvTime.tv_usec,
			maFnStats[ i ].tvChildTime.tv_sec, maFnStats[ i ].tvChildTime.tv_usec );
		fflush( pfile );
	}
	fclose( pfile );
}
